module Lebedev

use Constants
implicit none

! allowed values for integrationpoints: 6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302, 350, 434, 4802
integer, parameter		:: integrationpoints = 110

contains

! double precision function LebedevIntegration(integrand)
! implicit none
! ! integrand this is the function to be integrated
! ! and the programm calling this subroutine needs an "external integrand" statement
! ! the function has to be given in Cartesian coordinates
! double precision 	:: integral_value,r1,r2,r3,weight
! double precision	:: integrand
! integer			:: j
! integral_value = 0
! do j=1,integrationpoints
! call lebedevpoint(j,r1,r2,r3,weight)
! ! calculate function values of a sample function
! integral_value = integral_value + integrand(r1,r2,r3)*weight
! end do
! LebedevIntegration = integral_value
! end function LebedevIntegration


subroutine lebedevarray(theta_array,phi_array,weight_array)
! This subroutine returns an array of theta-values, phi-values and weights for the
! Lebedev-Laikov integration with 434 points on the unit sphere.
use SpinSphericals
implicit none
integer				:: j

double precision		:: theta_array(integrationpoints),phi_array(integrationpoints), weight_array(integrationpoints)
double precision		:: r1_current,r2_current,r3_current,weight_current,r,theta,phi

theta_array(:)  = 0.0d0
phi_array(:)    = 0.0d0
weight_array(:) = 0.0d0

do j=1,integrationpoints
    call lebedevpoint(j,r1_current,r2_current,r3_current,weight_current)
    call Cartesian2Spherical(r1_current,r2_current,r3_current,r,theta,phi)
    theta_array(j)  = theta
    phi_array(j)    = phi
    weight_array(j) = weight_current
end do
end subroutine lebedevarray


subroutine lebedevpoint(j,r1_current,r2_current,r3_current,weight_current)
! This subroutine returns the cartesian coordinates and the corresponding
! weights for a Lebedev-Laikov integration with 434 points on the unit sphere.
! j: 			running index from 1 to 434
! r1_current:		x-coordinate for a given point numbered by j
! r2_current:		y-coordinate for a given point numbered by j
! r3_current:		z-coordinate for a given point numbered by j
! weight_current:	weight for a given point numbered by j
!
! to calculate an integral of a function f(x,y,z) on the unit sphere
! sum up the function values at the points times the weights for j=0..434 
! 
! For test calculations compare to the Maple Worksheet CalculateSurfaceIntegral.mws


implicit none
integer				:: j

double precision, allocatable	:: r1(:),r2(:),r3(:), weight(:)
double precision		:: r1_current,r2_current,r3_current,weight_current

allocate(r1(integrationpoints))
allocate(r2(integrationpoints))
allocate(r3(integrationpoints))
allocate(weight(integrationpoints))

if(integrationpoints==4802) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.999454288285925196d0, &
      -0.999454288285925196d0, &
       0.999454288285925196d0, &
      -0.999454288285925196d0, &
       0.999454288285925196d0, &
      -0.999454288285925196d0, &
       0.999454288285925196d0, &
      -0.999454288285925196d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.998103351050866117d0, &
      -0.998103351050866117d0, &
       0.998103351050866117d0, &
      -0.998103351050866117d0, &
       0.998103351050866117d0, &
      -0.998103351050866117d0, &
       0.998103351050866117d0, &
      -0.998103351050866117d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.995845037809518829d0, &
      -0.995845037809518829d0, &
       0.995845037809518829d0, &
      -0.995845037809518829d0, &
       0.995845037809518829d0, &
      -0.995845037809518829d0, &
       0.995845037809518829d0, &
      -0.995845037809518829d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.991859768220476723d0, &
      -0.991859768220476723d0, &
       0.991859768220476723d0, &
      -0.991859768220476723d0, &
       0.991859768220476723d0, &
      -0.991859768220476723d0, &
       0.991859768220476723d0, &
      -0.991859768220476723d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.985574890876411480d0, &
      -0.985574890876411480d0, &
       0.985574890876411480d0, &
      -0.985574890876411480d0, &
       0.985574890876411480d0, &
      -0.985574890876411480d0, &
       0.985574890876411480d0, &
      -0.985574890876411480d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.976879895489590200d0, &
      -0.976879895489590200d0, &
       0.976879895489590200d0, &
      -0.976879895489590200d0, &
       0.976879895489590200d0, &
      -0.976879895489590200d0, &
       0.976879895489590200d0, &
      -0.976879895489590200d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.965703547196956058d0, &
      -0.965703547196956058d0, &
       0.965703547196956058d0, &
      -0.965703547196956058d0, &
       0.965703547196956058d0, &
      -0.965703547196956058d0, &
       0.965703547196956058d0, &
      -0.965703547196956058d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.951970830868824947d0, &
      -0.951970830868824947d0, &
       0.951970830868824947d0, &
      -0.951970830868824947d0, &
       0.951970830868824947d0, &
      -0.951970830868824947d0, &
       0.951970830868824947d0, &
      -0.951970830868824947d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.935616823535268760d0, &
      -0.935616823535268760d0, &
       0.935616823535268760d0, &
      -0.935616823535268760d0, &
       0.935616823535268760d0, &
      -0.935616823535268760d0, &
       0.935616823535268760d0, &
      -0.935616823535268760d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.916590817664864166d0, &
      -0.916590817664864166d0, &
       0.916590817664864166d0, &
      -0.916590817664864166d0, &
       0.916590817664864166d0, &
      -0.916590817664864166d0, &
       0.916590817664864166d0, &
      -0.916590817664864166d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.894857077778724364d0, &
      -0.894857077778724364d0, &
       0.894857077778724364d0, &
      -0.894857077778724364d0, &
       0.894857077778724364d0, &
      -0.894857077778724364d0, &
       0.894857077778724364d0, &
      -0.894857077778724364d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.870394385719328700d0, &
      -0.870394385719328700d0, &
       0.870394385719328700d0, &
      -0.870394385719328700d0, &
       0.870394385719328700d0, &
      -0.870394385719328700d0, &
       0.870394385719328700d0, &
      -0.870394385719328700d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.843194587148069274d0, &
      -0.843194587148069274d0, &
       0.843194587148069274d0, &
      -0.843194587148069274d0, &
       0.843194587148069274d0, &
      -0.843194587148069274d0, &
       0.843194587148069274d0, &
      -0.843194587148069274d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.813260572441469543d0, &
      -0.813260572441469543d0, &
       0.813260572441469543d0, &
      -0.813260572441469543d0, &
       0.813260572441469543d0, &
      -0.813260572441469543d0, &
       0.813260572441469543d0, &
      -0.813260572441469543d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.780604209137936267d0, &
      -0.780604209137936267d0, &
       0.780604209137936267d0, &
      -0.780604209137936267d0, &
       0.780604209137936267d0, &
      -0.780604209137936267d0, &
       0.780604209137936267d0, &
      -0.780604209137936267d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.745244567807627933d0, &
      -0.745244567807627933d0, &
       0.745244567807627933d0, &
      -0.745244567807627933d0, &
       0.745244567807627933d0, &
      -0.745244567807627933d0, &
       0.745244567807627933d0, &
      -0.745244567807627933d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.707206614402482003d0, &
      -0.707206614402482003d0, &
       0.707206614402482003d0, &
      -0.707206614402482003d0, &
       0.707206614402482003d0, &
      -0.707206614402482003d0, &
       0.707206614402482003d0, &
      -0.707206614402482003d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.666520468728925675d0, &
      -0.666520468728925675d0, &
       0.666520468728925675d0, &
      -0.666520468728925675d0, &
       0.666520468728925675d0, &
      -0.666520468728925675d0, &
       0.666520468728925675d0, &
      -0.666520468728925675d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.623221334571918639d0, &
      -0.623221334571918639d0, &
       0.623221334571918639d0, &
      -0.623221334571918639d0, &
       0.623221334571918639d0, &
      -0.623221334571918639d0, &
       0.623221334571918639d0, &
      -0.623221334571918639d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.528956067614563485d0, &
      -0.528956067614563485d0, &
       0.528956067614563485d0, &
      -0.528956067614563485d0, &
       0.528956067614563485d0, &
      -0.528956067614563485d0, &
       0.528956067614563485d0, &
      -0.528956067614563485d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.478098677043712394d0, &
      -0.478098677043712394d0, &
       0.478098677043712394d0, &
      -0.478098677043712394d0, &
       0.478098677043712394d0, &
      -0.478098677043712394d0, &
       0.478098677043712394d0, &
      -0.478098677043712394d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.424854689925054541d0, &
      -0.424854689925054541d0, &
       0.424854689925054541d0, &
      -0.424854689925054541d0, &
       0.424854689925054541d0, &
      -0.424854689925054541d0, &
       0.424854689925054541d0, &
      -0.424854689925054541d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.369325497276495829d0, &
      -0.369325497276495829d0, &
       0.369325497276495829d0, &
      -0.369325497276495829d0, &
       0.369325497276495829d0, &
      -0.369325497276495829d0, &
       0.369325497276495829d0, &
      -0.369325497276495829d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.311648435574098492d0, &
      -0.311648435574098492d0, &
       0.311648435574098492d0, &
      -0.311648435574098492d0, &
       0.311648435574098492d0, &
      -0.311648435574098492d0, &
       0.311648435574098492d0, &
      -0.311648435574098492d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.252010484558751358d0, &
      -0.252010484558751358d0, &
       0.252010484558751358d0, &
      -0.252010484558751358d0, &
       0.252010484558751358d0, &
      -0.252010484558751358d0, &
       0.252010484558751358d0, &
      -0.252010484558751358d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.190662629661763616d0, &
      -0.190662629661763616d0, &
       0.190662629661763616d0, &
      -0.190662629661763616d0, &
       0.190662629661763616d0, &
      -0.190662629661763616d0, &
       0.190662629661763616d0, &
      -0.190662629661763616d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.127931115176292354d0, &
      -0.127931115176292354d0, &
       0.127931115176292354d0, &
      -0.127931115176292354d0, &
       0.127931115176292354d0, &
      -0.127931115176292354d0, &
       0.127931115176292354d0, &
      -0.127931115176292354d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.064220087573159890d0, &
      -0.064220087573159890d0, &
       0.064220087573159890d0, &
      -0.064220087573159890d0, &
       0.064220087573159890d0, &
      -0.064220087573159890d0, &
       0.064220087573159890d0, &
      -0.064220087573159890d0, &
       0.045955576435858951d0, &
      -0.045955576435858951d0, &
       0.045955576435858951d0, &
      -0.045955576435858951d0, &
       0.998943484384601121d0, &
      -0.998943484384601121d0, &
       0.998943484384601121d0, &
      -0.998943484384601121d0, &
       0.045955576435858951d0, &
      -0.045955576435858951d0, &
       0.045955576435858951d0, &
      -0.045955576435858951d0, &
       0.998943484384601121d0, &
      -0.998943484384601121d0, &
       0.998943484384601121d0, &
      -0.998943484384601121d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.104931674243502304d0, &
      -0.104931674243502304d0, &
       0.104931674243502304d0, &
      -0.104931674243502304d0, &
       0.994479433543225988d0, &
      -0.994479433543225988d0, &
       0.994479433543225988d0, &
      -0.994479433543225988d0, &
       0.104931674243502304d0, &
      -0.104931674243502304d0, &
       0.104931674243502304d0, &
      -0.104931674243502304d0, &
       0.994479433543225988d0, &
      -0.994479433543225988d0, &
       0.994479433543225988d0, &
      -0.994479433543225988d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.177354887954927404d0, &
      -0.177354887954927404d0, &
       0.177354887954927404d0, &
      -0.177354887954927404d0, &
       0.984146962459619190d0, &
      -0.984146962459619190d0, &
       0.984146962459619190d0, &
      -0.984146962459619190d0, &
       0.177354887954927404d0, &
      -0.177354887954927404d0, &
       0.177354887954927404d0, &
      -0.177354887954927404d0, &
       0.984146962459619190d0, &
      -0.984146962459619190d0, &
       0.984146962459619190d0, &
      -0.984146962459619190d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.255907141123612680d0, &
      -0.255907141123612680d0, &
       0.255907141123612680d0, &
      -0.255907141123612680d0, &
       0.966701368118375703d0, &
      -0.966701368118375703d0, &
       0.966701368118375703d0, &
      -0.966701368118375703d0, &
       0.255907141123612680d0, &
      -0.255907141123612680d0, &
       0.255907141123612680d0, &
      -0.255907141123612680d0, &
       0.966701368118375703d0, &
      -0.966701368118375703d0, &
       0.966701368118375703d0, &
      -0.966701368118375703d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.335815683798589781d0, &
      -0.335815683798589781d0, &
       0.335815683798589781d0, &
      -0.335815683798589781d0, &
       0.941927718306922679d0, &
      -0.941927718306922679d0, &
       0.941927718306922679d0, &
      -0.941927718306922679d0, &
       0.335815683798589781d0, &
      -0.335815683798589781d0, &
       0.335815683798589781d0, &
      -0.335815683798589781d0, &
       0.941927718306922679d0, &
      -0.941927718306922679d0, &
       0.941927718306922679d0, &
      -0.941927718306922679d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.415583574376389320d0, &
      -0.415583574376389320d0, &
       0.415583574376389320d0, &
      -0.415583574376389320d0, &
       0.909554997077441252d0, &
      -0.909554997077441252d0, &
       0.909554997077441252d0, &
      -0.909554997077441252d0, &
       0.415583574376389320d0, &
      -0.415583574376389320d0, &
       0.415583574376389320d0, &
      -0.415583574376389320d0, &
       0.909554997077441252d0, &
      -0.909554997077441252d0, &
       0.909554997077441252d0, &
      -0.909554997077441252d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.493789429616747211d0, &
      -0.493789429616747211d0, &
       0.493789429616747211d0, &
      -0.493789429616747211d0, &
       0.869581508082346133d0, &
      -0.869581508082346133d0, &
       0.869581508082346133d0, &
      -0.869581508082346133d0, &
       0.493789429616747211d0, &
      -0.493789429616747211d0, &
       0.493789429616747211d0, &
      -0.493789429616747211d0, &
       0.869581508082346133d0, &
      -0.869581508082346133d0, &
       0.869581508082346133d0, &
      -0.869581508082346133d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.569156969479331587d0, &
      -0.569156969479331587d0, &
       0.569156969479331587d0, &
      -0.569156969479331587d0, &
       0.822228887897465732d0, &
      -0.822228887897465732d0, &
       0.822228887897465732d0, &
      -0.822228887897465732d0, &
       0.569156969479331587d0, &
      -0.569156969479331587d0, &
       0.569156969479331587d0, &
      -0.569156969479331587d0, &
       0.822228887897465732d0, &
      -0.822228887897465732d0, &
       0.822228887897465732d0, &
      -0.822228887897465732d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.640584085489425137d0, &
      -0.640584085489425137d0, &
       0.640584085489425137d0, &
      -0.640584085489425137d0, &
       0.767888031823440231d0, &
      -0.767888031823440231d0, &
       0.767888031823440231d0, &
      -0.767888031823440231d0, &
       0.640584085489425137d0, &
      -0.640584085489425137d0, &
       0.640584085489425137d0, &
      -0.640584085489425137d0, &
       0.767888031823440231d0, &
      -0.767888031823440231d0, &
       0.767888031823440231d0, &
      -0.767888031823440231d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.999454288285925196d0, &
       0.999454288285925196d0, &
      -0.999454288285925196d0, &
      -0.999454288285925196d0, &
       0.999454288285925196d0, &
       0.999454288285925196d0, &
      -0.999454288285925196d0, &
      -0.999454288285925196d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.998103351050866117d0, &
       0.998103351050866117d0, &
      -0.998103351050866117d0, &
      -0.998103351050866117d0, &
       0.998103351050866117d0, &
       0.998103351050866117d0, &
      -0.998103351050866117d0, &
      -0.998103351050866117d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.995845037809518829d0, &
       0.995845037809518829d0, &
      -0.995845037809518829d0, &
      -0.995845037809518829d0, &
       0.995845037809518829d0, &
       0.995845037809518829d0, &
      -0.995845037809518829d0, &
      -0.995845037809518829d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.991859768220476723d0, &
       0.991859768220476723d0, &
      -0.991859768220476723d0, &
      -0.991859768220476723d0, &
       0.991859768220476723d0, &
       0.991859768220476723d0, &
      -0.991859768220476723d0, &
      -0.991859768220476723d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.985574890876411480d0, &
       0.985574890876411480d0, &
      -0.985574890876411480d0, &
      -0.985574890876411480d0, &
       0.985574890876411480d0, &
       0.985574890876411480d0, &
      -0.985574890876411480d0, &
      -0.985574890876411480d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.976879895489590200d0, &
       0.976879895489590200d0, &
      -0.976879895489590200d0, &
      -0.976879895489590200d0, &
       0.976879895489590200d0, &
       0.976879895489590200d0, &
      -0.976879895489590200d0, &
      -0.976879895489590200d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.965703547196956058d0, &
       0.965703547196956058d0, &
      -0.965703547196956058d0, &
      -0.965703547196956058d0, &
       0.965703547196956058d0, &
       0.965703547196956058d0, &
      -0.965703547196956058d0, &
      -0.965703547196956058d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.951970830868824947d0, &
       0.951970830868824947d0, &
      -0.951970830868824947d0, &
      -0.951970830868824947d0, &
       0.951970830868824947d0, &
       0.951970830868824947d0, &
      -0.951970830868824947d0, &
      -0.951970830868824947d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.935616823535268760d0, &
       0.935616823535268760d0, &
      -0.935616823535268760d0, &
      -0.935616823535268760d0, &
       0.935616823535268760d0, &
       0.935616823535268760d0, &
      -0.935616823535268760d0, &
      -0.935616823535268760d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.916590817664864166d0, &
       0.916590817664864166d0, &
      -0.916590817664864166d0, &
      -0.916590817664864166d0, &
       0.916590817664864166d0, &
       0.916590817664864166d0, &
      -0.916590817664864166d0, &
      -0.916590817664864166d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.894857077778724364d0, &
       0.894857077778724364d0, &
      -0.894857077778724364d0, &
      -0.894857077778724364d0, &
       0.894857077778724364d0, &
       0.894857077778724364d0, &
      -0.894857077778724364d0, &
      -0.894857077778724364d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.870394385719328700d0, &
       0.870394385719328700d0, &
      -0.870394385719328700d0, &
      -0.870394385719328700d0, &
       0.870394385719328700d0, &
       0.870394385719328700d0, &
      -0.870394385719328700d0, &
      -0.870394385719328700d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.843194587148069274d0, &
       0.843194587148069274d0, &
      -0.843194587148069274d0, &
      -0.843194587148069274d0, &
       0.843194587148069274d0, &
       0.843194587148069274d0, &
      -0.843194587148069274d0, &
      -0.843194587148069274d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.813260572441469543d0, &
       0.813260572441469543d0, &
      -0.813260572441469543d0, &
      -0.813260572441469543d0, &
       0.813260572441469543d0, &
       0.813260572441469543d0, &
      -0.813260572441469543d0, &
      -0.813260572441469543d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.780604209137936267d0, &
       0.780604209137936267d0, &
      -0.780604209137936267d0, &
      -0.780604209137936267d0, &
       0.780604209137936267d0, &
       0.780604209137936267d0, &
      -0.780604209137936267d0, &
      -0.780604209137936267d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.745244567807627933d0, &
       0.745244567807627933d0, &
      -0.745244567807627933d0, &
      -0.745244567807627933d0, &
       0.745244567807627933d0, &
       0.745244567807627933d0, &
      -0.745244567807627933d0, &
      -0.745244567807627933d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.707206614402482003d0, &
       0.707206614402482003d0, &
      -0.707206614402482003d0, &
      -0.707206614402482003d0, &
       0.707206614402482003d0, &
       0.707206614402482003d0, &
      -0.707206614402482003d0, &
      -0.707206614402482003d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.666520468728925675d0, &
       0.666520468728925675d0, &
      -0.666520468728925675d0, &
      -0.666520468728925675d0, &
       0.666520468728925675d0, &
       0.666520468728925675d0, &
      -0.666520468728925675d0, &
      -0.666520468728925675d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.623221334571918639d0, &
       0.623221334571918639d0, &
      -0.623221334571918639d0, &
      -0.623221334571918639d0, &
       0.623221334571918639d0, &
       0.623221334571918639d0, &
      -0.623221334571918639d0, &
      -0.623221334571918639d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.528956067614563485d0, &
       0.528956067614563485d0, &
      -0.528956067614563485d0, &
      -0.528956067614563485d0, &
       0.528956067614563485d0, &
       0.528956067614563485d0, &
      -0.528956067614563485d0, &
      -0.528956067614563485d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.478098677043712394d0, &
       0.478098677043712394d0, &
      -0.478098677043712394d0, &
      -0.478098677043712394d0, &
       0.478098677043712394d0, &
       0.478098677043712394d0, &
      -0.478098677043712394d0, &
      -0.478098677043712394d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.424854689925054541d0, &
       0.424854689925054541d0, &
      -0.424854689925054541d0, &
      -0.424854689925054541d0, &
       0.424854689925054541d0, &
       0.424854689925054541d0, &
      -0.424854689925054541d0, &
      -0.424854689925054541d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.369325497276495829d0, &
       0.369325497276495829d0, &
      -0.369325497276495829d0, &
      -0.369325497276495829d0, &
       0.369325497276495829d0, &
       0.369325497276495829d0, &
      -0.369325497276495829d0, &
      -0.369325497276495829d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.311648435574098492d0, &
       0.311648435574098492d0, &
      -0.311648435574098492d0, &
      -0.311648435574098492d0, &
       0.311648435574098492d0, &
       0.311648435574098492d0, &
      -0.311648435574098492d0, &
      -0.311648435574098492d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.252010484558751358d0, &
       0.252010484558751358d0, &
      -0.252010484558751358d0, &
      -0.252010484558751358d0, &
       0.252010484558751358d0, &
       0.252010484558751358d0, &
      -0.252010484558751358d0, &
      -0.252010484558751358d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.190662629661763616d0, &
       0.190662629661763616d0, &
      -0.190662629661763616d0, &
      -0.190662629661763616d0, &
       0.190662629661763616d0, &
       0.190662629661763616d0, &
      -0.190662629661763616d0, &
      -0.190662629661763616d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.127931115176292354d0, &
       0.127931115176292354d0, &
      -0.127931115176292354d0, &
      -0.127931115176292354d0, &
       0.127931115176292354d0, &
       0.127931115176292354d0, &
      -0.127931115176292354d0, &
      -0.127931115176292354d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.064220087573159890d0, &
       0.064220087573159890d0, &
      -0.064220087573159890d0, &
      -0.064220087573159890d0, &
       0.064220087573159890d0, &
       0.064220087573159890d0, &
      -0.064220087573159890d0, &
      -0.064220087573159890d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.998943484384601121d0, &
       0.998943484384601121d0, &
      -0.998943484384601121d0, &
      -0.998943484384601121d0, &
       0.045955576435858951d0, &
       0.045955576435858951d0, &
      -0.045955576435858951d0, &
      -0.045955576435858951d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.045955576435858951d0, &
      -0.045955576435858951d0, &
       0.045955576435858951d0, &
      -0.045955576435858951d0, &
       0.998943484384601121d0, &
      -0.998943484384601121d0, &
       0.998943484384601121d0, &
      -0.998943484384601121d0, &
       0.994479433543225988d0, &
       0.994479433543225988d0, &
      -0.994479433543225988d0, &
      -0.994479433543225988d0, &
       0.104931674243502304d0, &
       0.104931674243502304d0, &
      -0.104931674243502304d0, &
      -0.104931674243502304d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.104931674243502304d0, &
      -0.104931674243502304d0, &
       0.104931674243502304d0, &
      -0.104931674243502304d0, &
       0.994479433543225988d0, &
      -0.994479433543225988d0, &
       0.994479433543225988d0, &
      -0.994479433543225988d0, &
       0.984146962459619190d0, &
       0.984146962459619190d0, &
      -0.984146962459619190d0, &
      -0.984146962459619190d0, &
       0.177354887954927404d0, &
       0.177354887954927404d0, &
      -0.177354887954927404d0, &
      -0.177354887954927404d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.177354887954927404d0, &
      -0.177354887954927404d0, &
       0.177354887954927404d0, &
      -0.177354887954927404d0, &
       0.984146962459619190d0, &
      -0.984146962459619190d0, &
       0.984146962459619190d0, &
      -0.984146962459619190d0, &
       0.966701368118375703d0, &
       0.966701368118375703d0, &
      -0.966701368118375703d0, &
      -0.966701368118375703d0, &
       0.255907141123612680d0, &
       0.255907141123612680d0, &
      -0.255907141123612680d0, &
      -0.255907141123612680d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.255907141123612680d0, &
      -0.255907141123612680d0, &
       0.255907141123612680d0, &
      -0.255907141123612680d0, &
       0.966701368118375703d0, &
      -0.966701368118375703d0, &
       0.966701368118375703d0, &
      -0.966701368118375703d0, &
       0.941927718306922679d0, &
       0.941927718306922679d0, &
      -0.941927718306922679d0, &
      -0.941927718306922679d0, &
       0.335815683798589781d0, &
       0.335815683798589781d0, &
      -0.335815683798589781d0, &
      -0.335815683798589781d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.335815683798589781d0, &
      -0.335815683798589781d0, &
       0.335815683798589781d0, &
      -0.335815683798589781d0, &
       0.941927718306922679d0, &
      -0.941927718306922679d0, &
       0.941927718306922679d0, &
      -0.941927718306922679d0, &
       0.909554997077441252d0, &
       0.909554997077441252d0, &
      -0.909554997077441252d0, &
      -0.909554997077441252d0, &
       0.415583574376389320d0, &
       0.415583574376389320d0, &
      -0.415583574376389320d0, &
      -0.415583574376389320d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.415583574376389320d0, &
      -0.415583574376389320d0, &
       0.415583574376389320d0, &
      -0.415583574376389320d0, &
       0.909554997077441252d0, &
      -0.909554997077441252d0, &
       0.909554997077441252d0, &
      -0.909554997077441252d0, &
       0.869581508082346133d0, &
       0.869581508082346133d0, &
      -0.869581508082346133d0, &
      -0.869581508082346133d0, &
       0.493789429616747211d0, &
       0.493789429616747211d0, &
      -0.493789429616747211d0, &
      -0.493789429616747211d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.493789429616747211d0, &
      -0.493789429616747211d0, &
       0.493789429616747211d0, &
      -0.493789429616747211d0, &
       0.869581508082346133d0, &
      -0.869581508082346133d0, &
       0.869581508082346133d0, &
      -0.869581508082346133d0, &
       0.822228887897465732d0, &
       0.822228887897465732d0, &
      -0.822228887897465732d0, &
      -0.822228887897465732d0, &
       0.569156969479331587d0, &
       0.569156969479331587d0, &
      -0.569156969479331587d0, &
      -0.569156969479331587d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.569156969479331587d0, &
      -0.569156969479331587d0, &
       0.569156969479331587d0, &
      -0.569156969479331587d0, &
       0.822228887897465732d0, &
      -0.822228887897465732d0, &
       0.822228887897465732d0, &
      -0.822228887897465732d0, &
       0.767888031823440231d0, &
       0.767888031823440231d0, &
      -0.767888031823440231d0, &
      -0.767888031823440231d0, &
       0.640584085489425137d0, &
       0.640584085489425137d0, &
      -0.640584085489425137d0, &
      -0.640584085489425137d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.640584085489425137d0, &
      -0.640584085489425137d0, &
       0.640584085489425137d0, &
      -0.640584085489425137d0, &
       0.767888031823440231d0, &
      -0.767888031823440231d0, &
       0.767888031823440231d0, &
      -0.767888031823440231d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.999454288285925196d0, &
       0.999454288285925196d0, &
       0.999454288285925196d0, &
       0.999454288285925196d0, &
      -0.999454288285925196d0, &
      -0.999454288285925196d0, &
      -0.999454288285925196d0, &
      -0.999454288285925196d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
       0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
      -0.023357286088870641d0, &
       0.998103351050866117d0, &
       0.998103351050866117d0, &
       0.998103351050866117d0, &
       0.998103351050866117d0, &
      -0.998103351050866117d0, &
      -0.998103351050866117d0, &
      -0.998103351050866117d0, &
      -0.998103351050866117d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
       0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
      -0.043529878365506529d0, &
       0.995845037809518829d0, &
       0.995845037809518829d0, &
       0.995845037809518829d0, &
       0.995845037809518829d0, &
      -0.995845037809518829d0, &
      -0.995845037809518829d0, &
      -0.995845037809518829d0, &
      -0.995845037809518829d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
       0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
      -0.064392005210888004d0, &
       0.991859768220476723d0, &
       0.991859768220476723d0, &
       0.991859768220476723d0, &
       0.991859768220476723d0, &
      -0.991859768220476723d0, &
      -0.991859768220476723d0, &
      -0.991859768220476723d0, &
      -0.991859768220476723d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
       0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
      -0.090039436319931804d0, &
       0.985574890876411480d0, &
       0.985574890876411480d0, &
       0.985574890876411480d0, &
       0.985574890876411480d0, &
      -0.985574890876411480d0, &
      -0.985574890876411480d0, &
      -0.985574890876411480d0, &
      -0.985574890876411480d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
       0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
      -0.119670661554847305d0, &
       0.976879895489590200d0, &
       0.976879895489590200d0, &
       0.976879895489590200d0, &
       0.976879895489590200d0, &
      -0.976879895489590200d0, &
      -0.976879895489590200d0, &
      -0.976879895489590200d0, &
      -0.976879895489590200d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
       0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
      -0.151171541283813410d0, &
       0.965703547196956058d0, &
       0.965703547196956058d0, &
       0.965703547196956058d0, &
       0.965703547196956058d0, &
      -0.965703547196956058d0, &
      -0.965703547196956058d0, &
      -0.965703547196956058d0, &
      -0.965703547196956058d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
       0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
      -0.183598282850380101d0, &
       0.951970830868824947d0, &
       0.951970830868824947d0, &
       0.951970830868824947d0, &
       0.951970830868824947d0, &
      -0.951970830868824947d0, &
      -0.951970830868824947d0, &
      -0.951970830868824947d0, &
      -0.951970830868824947d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
       0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
      -0.216508125915540500d0, &
       0.935616823535268760d0, &
       0.935616823535268760d0, &
       0.935616823535268760d0, &
       0.935616823535268760d0, &
      -0.935616823535268760d0, &
      -0.935616823535268760d0, &
      -0.935616823535268760d0, &
      -0.935616823535268760d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
       0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
      -0.249620872041756287d0, &
       0.916590817664864166d0, &
       0.916590817664864166d0, &
       0.916590817664864166d0, &
       0.916590817664864166d0, &
      -0.916590817664864166d0, &
      -0.916590817664864166d0, &
      -0.916590817664864166d0, &
      -0.916590817664864166d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
       0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
      -0.282720067356790006d0, &
       0.894857077778724364d0, &
       0.894857077778724364d0, &
       0.894857077778724364d0, &
       0.894857077778724364d0, &
      -0.894857077778724364d0, &
      -0.894857077778724364d0, &
      -0.894857077778724364d0, &
      -0.894857077778724364d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
       0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
      -0.315619082399434592d0, &
       0.870394385719328700d0, &
       0.870394385719328700d0, &
       0.870394385719328700d0, &
       0.870394385719328700d0, &
      -0.870394385719328700d0, &
      -0.870394385719328700d0, &
      -0.870394385719328700d0, &
      -0.870394385719328700d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
       0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
      -0.348147679374911478d0, &
       0.843194587148069274d0, &
       0.843194587148069274d0, &
       0.843194587148069274d0, &
       0.843194587148069274d0, &
      -0.843194587148069274d0, &
      -0.843194587148069274d0, &
      -0.843194587148069274d0, &
      -0.843194587148069274d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
       0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
      -0.380146608694722610d0, &
       0.813260572441469543d0, &
       0.813260572441469543d0, &
       0.813260572441469543d0, &
       0.813260572441469543d0, &
      -0.813260572441469543d0, &
      -0.813260572441469543d0, &
      -0.813260572441469543d0, &
      -0.813260572441469543d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
       0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
      -0.411465211963401079d0, &
       0.780604209137936267d0, &
       0.780604209137936267d0, &
       0.780604209137936267d0, &
       0.780604209137936267d0, &
      -0.780604209137936267d0, &
      -0.780604209137936267d0, &
      -0.780604209137936267d0, &
      -0.780604209137936267d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
       0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
      -0.441959878651975113d0, &
       0.745244567807627933d0, &
       0.745244567807627933d0, &
       0.745244567807627933d0, &
       0.745244567807627933d0, &
      -0.745244567807627933d0, &
      -0.745244567807627933d0, &
      -0.745244567807627933d0, &
      -0.745244567807627933d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
       0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
      -0.471492594932954312d0, &
       0.707206614402482003d0, &
       0.707206614402482003d0, &
       0.707206614402482003d0, &
       0.707206614402482003d0, &
      -0.707206614402482003d0, &
      -0.707206614402482003d0, &
      -0.707206614402482003d0, &
      -0.707206614402482003d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
       0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
      -0.499929397287946609d0, &
       0.666520468728925675d0, &
       0.666520468728925675d0, &
       0.666520468728925675d0, &
       0.666520468728925675d0, &
      -0.666520468728925675d0, &
      -0.666520468728925675d0, &
      -0.666520468728925675d0, &
      -0.666520468728925675d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
       0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
      -0.527138722143124849d0, &
       0.623221334571918639d0, &
       0.623221334571918639d0, &
       0.623221334571918639d0, &
       0.623221334571918639d0, &
      -0.623221334571918639d0, &
      -0.623221334571918639d0, &
      -0.623221334571918639d0, &
      -0.623221334571918639d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
       0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
      -0.552989678083776082d0, &
       0.528956067614563485d0, &
       0.528956067614563485d0, &
       0.528956067614563485d0, &
       0.528956067614563485d0, &
      -0.528956067614563485d0, &
      -0.528956067614563485d0, &
      -0.528956067614563485d0, &
      -0.528956067614563485d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
       0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
      -0.600085609948171173d0, &
       0.478098677043712394d0, &
       0.478098677043712394d0, &
       0.478098677043712394d0, &
       0.478098677043712394d0, &
      -0.478098677043712394d0, &
      -0.478098677043712394d0, &
      -0.478098677043712394d0, &
      -0.478098677043712394d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
       0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
      -0.621056219278517507d0, &
       0.424854689925054541d0, &
       0.424854689925054541d0, &
       0.424854689925054541d0, &
       0.424854689925054541d0, &
      -0.424854689925054541d0, &
      -0.424854689925054541d0, &
      -0.424854689925054541d0, &
      -0.424854689925054541d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
       0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
      -0.640116587993423969d0, &
       0.369325497276495829d0, &
       0.369325497276495829d0, &
       0.369325497276495829d0, &
       0.369325497276495829d0, &
      -0.369325497276495829d0, &
      -0.369325497276495829d0, &
      -0.369325497276495829d0, &
      -0.369325497276495829d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
       0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
      -0.657114402924433350d0, &
       0.311648435574098492d0, &
       0.311648435574098492d0, &
       0.311648435574098492d0, &
       0.311648435574098492d0, &
      -0.311648435574098492d0, &
      -0.311648435574098492d0, &
      -0.311648435574098492d0, &
      -0.311648435574098492d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
       0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
      -0.671891082171886311d0, &
       0.252010484558751358d0, &
       0.252010484558751358d0, &
       0.252010484558751358d0, &
       0.252010484558751358d0, &
      -0.252010484558751358d0, &
      -0.252010484558751358d0, &
      -0.252010484558751358d0, &
      -0.252010484558751358d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
       0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
      -0.684284559109901047d0, &
       0.190662629661763616d0, &
       0.190662629661763616d0, &
       0.190662629661763616d0, &
       0.190662629661763616d0, &
      -0.190662629661763616d0, &
      -0.190662629661763616d0, &
      -0.190662629661763616d0, &
      -0.190662629661763616d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
       0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
      -0.694135347626981591d0, &
       0.127931115176292354d0, &
       0.127931115176292354d0, &
       0.127931115176292354d0, &
       0.127931115176292354d0, &
      -0.127931115176292354d0, &
      -0.127931115176292354d0, &
      -0.127931115176292354d0, &
      -0.127931115176292354d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
       0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
      -0.701296524221299067d0, &
       0.064220087573159890d0, &
       0.064220087573159890d0, &
       0.064220087573159890d0, &
       0.064220087573159890d0, &
      -0.064220087573159890d0, &
      -0.064220087573159890d0, &
      -0.064220087573159890d0, &
      -0.064220087573159890d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
       0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
      -0.705647142824264439d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.998943484384601121d0, &
       0.998943484384601121d0, &
      -0.998943484384601121d0, &
      -0.998943484384601121d0, &
       0.045955576435858951d0, &
       0.045955576435858951d0, &
      -0.045955576435858951d0, &
      -0.045955576435858951d0, &
       0.998943484384601121d0, &
       0.998943484384601121d0, &
      -0.998943484384601121d0, &
      -0.998943484384601121d0, &
       0.045955576435858951d0, &
       0.045955576435858951d0, &
      -0.045955576435858951d0, &
      -0.045955576435858951d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.994479433543225988d0, &
       0.994479433543225988d0, &
      -0.994479433543225988d0, &
      -0.994479433543225988d0, &
       0.104931674243502304d0, &
       0.104931674243502304d0, &
      -0.104931674243502304d0, &
      -0.104931674243502304d0, &
       0.994479433543225988d0, &
       0.994479433543225988d0, &
      -0.994479433543225988d0, &
      -0.994479433543225988d0, &
       0.104931674243502304d0, &
       0.104931674243502304d0, &
      -0.104931674243502304d0, &
      -0.104931674243502304d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.984146962459619190d0, &
       0.984146962459619190d0, &
      -0.984146962459619190d0, &
      -0.984146962459619190d0, &
       0.177354887954927404d0, &
       0.177354887954927404d0, &
      -0.177354887954927404d0, &
      -0.177354887954927404d0, &
       0.984146962459619190d0, &
       0.984146962459619190d0, &
      -0.984146962459619190d0, &
      -0.984146962459619190d0, &
       0.177354887954927404d0, &
       0.177354887954927404d0, &
      -0.177354887954927404d0, &
      -0.177354887954927404d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.966701368118375703d0, &
       0.966701368118375703d0, &
      -0.966701368118375703d0, &
      -0.966701368118375703d0, &
       0.255907141123612680d0, &
       0.255907141123612680d0, &
      -0.255907141123612680d0, &
      -0.255907141123612680d0, &
       0.966701368118375703d0, &
       0.966701368118375703d0, &
      -0.966701368118375703d0, &
      -0.966701368118375703d0, &
       0.255907141123612680d0, &
       0.255907141123612680d0, &
      -0.255907141123612680d0, &
      -0.255907141123612680d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.941927718306922679d0, &
       0.941927718306922679d0, &
      -0.941927718306922679d0, &
      -0.941927718306922679d0, &
       0.335815683798589781d0, &
       0.335815683798589781d0, &
      -0.335815683798589781d0, &
      -0.335815683798589781d0, &
       0.941927718306922679d0, &
       0.941927718306922679d0, &
      -0.941927718306922679d0, &
      -0.941927718306922679d0, &
       0.335815683798589781d0, &
       0.335815683798589781d0, &
      -0.335815683798589781d0, &
      -0.335815683798589781d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.909554997077441252d0, &
       0.909554997077441252d0, &
      -0.909554997077441252d0, &
      -0.909554997077441252d0, &
       0.415583574376389320d0, &
       0.415583574376389320d0, &
      -0.415583574376389320d0, &
      -0.415583574376389320d0, &
       0.909554997077441252d0, &
       0.909554997077441252d0, &
      -0.909554997077441252d0, &
      -0.909554997077441252d0, &
       0.415583574376389320d0, &
       0.415583574376389320d0, &
      -0.415583574376389320d0, &
      -0.415583574376389320d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.869581508082346133d0, &
       0.869581508082346133d0, &
      -0.869581508082346133d0, &
      -0.869581508082346133d0, &
       0.493789429616747211d0, &
       0.493789429616747211d0, &
      -0.493789429616747211d0, &
      -0.493789429616747211d0, &
       0.869581508082346133d0, &
       0.869581508082346133d0, &
      -0.869581508082346133d0, &
      -0.869581508082346133d0, &
       0.493789429616747211d0, &
       0.493789429616747211d0, &
      -0.493789429616747211d0, &
      -0.493789429616747211d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.822228887897465732d0, &
       0.822228887897465732d0, &
      -0.822228887897465732d0, &
      -0.822228887897465732d0, &
       0.569156969479331587d0, &
       0.569156969479331587d0, &
      -0.569156969479331587d0, &
      -0.569156969479331587d0, &
       0.822228887897465732d0, &
       0.822228887897465732d0, &
      -0.822228887897465732d0, &
      -0.822228887897465732d0, &
       0.569156969479331587d0, &
       0.569156969479331587d0, &
      -0.569156969479331587d0, &
      -0.569156969479331587d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.767888031823440231d0, &
       0.767888031823440231d0, &
      -0.767888031823440231d0, &
      -0.767888031823440231d0, &
       0.640584085489425137d0, &
       0.640584085489425137d0, &
      -0.640584085489425137d0, &
      -0.640584085489425137d0, &
       0.767888031823440231d0, &
       0.767888031823440231d0, &
      -0.767888031823440231d0, &
      -0.767888031823440231d0, &
       0.640584085489425137d0, &
       0.640584085489425137d0, &
      -0.640584085489425137d0, &
      -0.640584085489425137d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
       0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
      -0.997060981246074229d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
       0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
      -0.021778440814860672d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
       0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
      -0.073451338941433478d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
       0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
      -0.993828299383836211d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
       0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
      -0.045903621857751882d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
       0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
      -0.100985983404493093d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
       0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
      -0.988533750550789958d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
       0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
      -0.072550630956908765d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
       0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
      -0.132428961974875803d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
       0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
      -0.980955733643047778d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
       0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
      -0.101782545196068400d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
       0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
      -0.165427210960712695d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
       0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
      -0.970976271244223144d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
       0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
      -0.132565232098036390d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
       0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
      -0.199076718677646092d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
       0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
      -0.958497965581324274d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
       0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
      -0.164276537449676491d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
       0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
      -0.233012594552327812d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
       0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
      -0.943440661245666501d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
       0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
      -0.196536037433788890d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
       0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
      -0.267008061110828698d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
       0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
      -0.925742804365006666d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
       0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
      -0.229072677054223794d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
       0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
      -0.300875337629431583d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
       0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
      -0.905359869544664631d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
       0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
      -0.261664549537082292d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
       0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
      -0.334447559616786017d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
       0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
      -0.882262945014676059d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
       0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
      -0.294115072884314099d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
       0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
      -0.367570972407078589d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
       0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
      -0.856437239545155271d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
       0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
      -0.326244040091906606d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
       0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
      -0.400100088758781225d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
       0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
      -0.827880145771347720d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
       0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
      -0.357883535061191582d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
       0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
      -0.431895635043602788d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
       0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
      -0.796599097732506034d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
       0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
      -0.388875185404367774d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
       0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
      -0.462823905679553116d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
       0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
      -0.762609588779783465d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
       0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
      -0.419067800322284023d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
       0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
      -0.492756322977363603d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
       0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
      -0.725933587178771988d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
       0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
      -0.448315183688385188d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
       0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
      -0.521568713670796935d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
       0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
      -0.686598474751890842d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
       0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
      -0.476474067608788021d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
       0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
      -0.549140234698490493d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
       0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
      -0.644636604665279722d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
       0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
      -0.503402131099827677d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
       0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
      -0.575352016012607548d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
       0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
      -0.990016345411100951d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
       0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
      -0.024354365103728060d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
       0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
      -0.138832635641775393d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
       0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
      -0.983348996654077601d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
       0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
      -0.051188970573426522d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
       0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
      -0.174368690053724407d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
       0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
      -0.974416497213840160d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
       0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
      -0.080146950485396343d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
       0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
      -0.209973703795026806d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
       0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
      -0.963089718590104282d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
       0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
      -0.110511787415569898d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
       0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
      -0.245449259090854799d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
       0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
      -0.949257269281060356d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
       0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
      -0.141795053157096590d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
       0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
      -0.280721925786427795d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
       0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
      -0.932837339156021517d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
       0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
      -0.173660494571959700d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
       0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
      -0.315684227197584211d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
       0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
      -0.913772813131386918d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
       0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
      -0.205846632469398105d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
       0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
      -0.350209094517775210d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
       0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
      -0.892025463674392616d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
       0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
      -0.238128426119591896d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
       0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
      -0.384168484951968592d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
       0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
      -0.867572690240884015d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
       0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
      -0.270303127042256897d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
       0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
      -0.417437236790601573d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
       0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
      -0.840405314892748834d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
       0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
      -0.302184568309130874d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
       0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
      -0.449892646501189208d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
       0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
      -0.810525412385808330d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
       0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
      -0.333599335516572004d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
       0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
      -0.481414622980770091d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
       0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
      -0.777944155380252522d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
       0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
      -0.364383373551823209d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
       0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
      -0.511886362573470088d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
       0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
      -0.742679936390977957d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
       0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
      -0.394378954195817921d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
       0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
      -0.541194745511914377d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
       0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
      -0.704756954869755203d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
       0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
      -0.423432014440354221d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
       0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
      -0.569230150035724636d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
       0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
      -0.664204382257133741d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
       0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
      -0.451389794741925976d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
       0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
      -0.595885720413957598d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
       0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
      -0.976107620825866329d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
       0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
      -0.026812257554444909d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
       0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
      -0.215627028478576593d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
       0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
      -0.965806234934322494d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
       0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
      -0.055574957478056138d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
       0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
      -0.253238505490970989d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
       0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
      -0.953104285740430712d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
       0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
      -0.085693680629502486d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
       0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
      -0.290256461777153696d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
       0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
      -0.937891763857101934d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
       0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
      -0.116736745032413505d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
       0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
      -0.326697982314325619d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
       0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
      -0.920093697847417569d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
       0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
      -0.148386199400330404d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
       0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
      -0.362503962749361419d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
       0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
      -0.899660673409398925d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
       0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
      -0.180382150301140493d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
       0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
      -0.397583893754869910d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
       0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
      -0.876561392753544277d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
       0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
      -0.212496296566642401d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
       0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
      -0.431839609900977417d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
       0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
      -0.850779150445461374d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
       0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
      -0.244522183780591290d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
       0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
      -0.465170655573274194d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
       0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
      -0.822309661990015983d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
       0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
      -0.276270122432298704d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
       0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
      -0.497475264962096875d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
       0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
      -0.791158932633878820d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
       0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
      -0.307562777521132813d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
       0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
      -0.528651757962751656d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
       0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
      -0.757341153859447780d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
       0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
      -0.338231108982687712d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
       0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
      -0.558600119573189446d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
       0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
      -0.720876922397960418d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
       0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
      -0.368110883474139894d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
       0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
      -0.587222990202131867d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
       0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
      -0.681791978323931902d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
       0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
      -0.397039744687283924d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
       0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
      -0.614425861623512315d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
       0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
      -0.955015079753669061d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
       0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
      -0.028674995387504409d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
       0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
      -0.295167650806486104d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
       0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
      -0.940919150218964218d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
       0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
      -0.058678793419035098d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
       0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
      -0.333508548547272521d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
       0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
      -0.924316712790598571d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
       0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
      -0.089610992050222840d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
       0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
      -0.370956176063638121d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
       0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
      -0.905144139712566687d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
       0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
      -0.121162792762629701d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
       0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
      -0.407472286166749798d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
       0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
      -0.883360539415936397d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
       0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
      -0.153074890355489796d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
       0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
      -0.442992364883911693d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
       0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
      -0.858941107233257450d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
       0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
      -0.185117643672187693d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
       0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
      -0.477442805272173587d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
       0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
      -0.831874334446776387d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
       0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
      -0.217082910765817899d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
       0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
      -0.510744653953590366d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
       0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
      -0.802160146656367812d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
       0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
      -0.248778668902627093d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
       0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
      -0.542815137054293517d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
       0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
      -0.769807832073252540d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
       0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
      -0.280023995279501614d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
       0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
      -0.573569929255696387d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
       0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
      -0.734833952507771393d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
       0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
      -0.310644570287811894d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
       0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
      -0.602925379456286548d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
       0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
      -0.697260626896040292d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
       0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
      -0.340468950084119426d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
       0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
      -0.630799898707314455d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
       0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
      -0.926432793705799007d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
       0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
      -0.029971450981844790d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
       0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
      -0.375265227369271925d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
       0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
      -0.908449876706498127d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
       0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
      -0.060867258986780109d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
       0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
      -0.413538387934402785d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
       0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
      -0.887926653751518358d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
       0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
      -0.092388495484356431d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
       0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
      -0.450611388515390676d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
       0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
      -0.864829919538127490d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
       0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
      -0.124278660385185105d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
       0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
      -0.486440155460607204d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
       0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
      -0.839140641527672093d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
       0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
      -0.156308673148338600d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
       0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
      -0.520970807661170876d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
       0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
      -0.810851987517254957d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
       0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
      -0.188269650938850591d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
       0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
      -0.554142213583012189d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
       0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
      -0.779967648094571753d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
       0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
      -0.219967297912605914d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
       0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
      -0.585888091511381659d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
       0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
      -0.746499712899290957d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
       0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
      -0.251216548292486686d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
       0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
      -0.616139939060344366d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
       0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
      -0.710466539234928307d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
       0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
      -0.281836870187188782d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
       0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
      -0.644829648225509033d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
       0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
      -0.890221373804620297d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
       0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
      -0.030889704050603119d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
       0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
      -0.454479627491794824d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
       0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
      -0.868389986571093120d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
       0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
      -0.062409476776368350d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
       0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
      -0.491938907214662824d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
       0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
      -0.844034784705623875d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
       0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
      -0.094307061442803125d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
       0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
      -0.527931302698518290d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
       0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
      -0.817142334957513983d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
       0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
      -0.126354781877037387d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
       0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
      -0.562416992557113526d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
       0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
      -0.787710401936984494d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
       0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
      -0.158343078882259403d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
       0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
      -0.595348462709328730d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
       0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
      -0.755746263130129381d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
       0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
      -0.190074846255598812d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
       0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
      -0.626673071533918491d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
       0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
      -0.721264449529974705d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
       0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
      -0.221359951959256701d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
       0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
      -0.656336320427887143d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
       0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
      -0.846498153948142096d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
       0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
      -0.031525088115153740d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
       0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
      -0.531457471658569602d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
       0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
      -0.820952490109968847d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
       0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
      -0.063438652914655613d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
       0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
      -0.567461493229818537d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
       0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
      -0.792936959952257547d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
       0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
      -0.095515035042239516d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
       0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
      -0.601770600497026353d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
       0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
      -0.762454095635613172d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
       0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
      -0.127544009980119610d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
       0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
      -0.634347127026417823d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
       0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
      -0.729514684857185292d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
       0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
      -0.159325203767196000d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
       0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
      -0.665149459912780183d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
       0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
      -0.795571106970856978d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
       0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
      -0.031925383384961051d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
       0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
      -0.605018498600570442d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
       0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
      -0.766523634312228364d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
       0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
      -0.064028243539623053d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
       0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
      -0.639016355088040000d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
       0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
      -0.735094028059236182d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
       0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
      -0.096098050770029086d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
       0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
      -0.671119910708844802d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
       0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
      -0.737909076006943887d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
       0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0, &
      -0.032118531962732330d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
       0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0, &
      -0.674135442957227493d0/)
 
weight = (/&
       0.000096875218794207d0, &
       0.000096875218794207d0, &
       0.000096875218794207d0, &
       0.000096875218794207d0, &
       0.000096875218794207d0, &
       0.000096875218794207d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000230789789536792d0, &
       0.000229731085249856d0, &
       0.000229731085249856d0, &
       0.000229731085249856d0, &
       0.000229731085249856d0, &
       0.000229731085249856d0, &
       0.000229731085249856d0, &
       0.000229731085249856d0, &
       0.000229731085249856d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000073862659440019d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000082579776985422d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000097060447620576d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000130239384711700d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000154195700460097d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000170445977009220d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000182737489094291d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000192636081743611d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000200801023949483d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000207563598320917d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000213130663869091d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000217656232993734d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000221268226299102d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000224079951566857d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000226195981618753d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000227715636880886d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000228735177212834d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229349081408408d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229650531237627d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229679383231876d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229578544384297d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229501793152910d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229505963818487d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229623234323736d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000229853017874077d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230157979028050d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230469040499651d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000230702799590710d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000093122746966711d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000119991938587693d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000159803913887769d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000182225376357490d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000198857959365504d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000211262010253331d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000220159488769901d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000226162259089504d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000229645845343571d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000100600699026700d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000122767668963588d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000146786428027012d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000164417891210123d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000177766489071896d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000188482566451669d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000197326924645385d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000204676777585533d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000210760012591804d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000215741636226683d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000219755781692072d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000222919261183544d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000225338511021277d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000227113710754877d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000228341409291752d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229116167313008d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000229531390857660d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000143820472135903d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000160773802549526d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000174148385352838d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000185191846751915d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000194462863807061d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000202249544627515d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000208746238243851d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000214107475481831d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000218464091374816d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000221930916522033d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000224612311834062d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000226606276691512d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228007295223080d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000228908202520258d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000229401269512003d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000172243448873695d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000183023742145509d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000192385534999763d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000200406786193627d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000207181729735426d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000212825083410210d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000217451371944010d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000221166183915021d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000224066525781310d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000226243951663262d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000227787455723187d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000228785431445499d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000229326849961558d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000191262820152983d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000199249967223870d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000206127553345403d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000211931821596857d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000216741658188265d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000220643073051660d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000223718693869952d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000226048007503288d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000227709888455854d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000228784571510967d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000229354726823629d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000205607383985253d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000211423586583188d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000216317562977055d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000220339215811165d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000223547317684784d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000226002414150124d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000227767592932918d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000228910211228483d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000229502795462512d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000216128158987999d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000220198047739510d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000223495206659317d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000226054009852084d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000227915798189999d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229129691856557d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000229753375253665d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000223492735646600d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000226128801298522d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000228081816092369d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000229377329518016d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000230052876733863d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000228189385506567d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000229572044484073d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230322764902675d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0, &
       0.000230483191322711d0/)

else if(integrationpoints==434) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.212646824707551862d0, &
      -0.212646824707551862d0, &
       0.212646824707551862d0, &
      -0.212646824707551862d0, &
       0.212646824707551862d0, &
      -0.212646824707551862d0, &
       0.212646824707551862d0, &
      -0.212646824707551862d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.967987158791472790d0, &
      -0.967987158791472790d0, &
       0.967987158791472790d0, &
      -0.967987158791472790d0, &
       0.967987158791472790d0, &
      -0.967987158791472790d0, &
       0.967987158791472790d0, &
      -0.967987158791472790d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.719016501040843470d0, &
      -0.719016501040843470d0, &
       0.719016501040843470d0, &
      -0.719016501040843470d0, &
       0.719016501040843470d0, &
      -0.719016501040843470d0, &
       0.719016501040843470d0, &
      -0.719016501040843470d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.407712664897769750d0, &
      -0.407712664897769750d0, &
       0.407712664897769750d0, &
      -0.407712664897769750d0, &
       0.407712664897769750d0, &
      -0.407712664897769750d0, &
       0.407712664897769750d0, &
      -0.407712664897769750d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.914472801120872480d0, &
      -0.914472801120872480d0, &
       0.914472801120872480d0, &
      -0.914472801120872480d0, &
       0.914472801120872480d0, &
      -0.914472801120872480d0, &
       0.914472801120872480d0, &
      -0.914472801120872480d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.994255912631277883d0, &
      -0.994255912631277883d0, &
       0.994255912631277883d0, &
      -0.994255912631277883d0, &
       0.994255912631277883d0, &
      -0.994255912631277883d0, &
       0.994255912631277883d0, &
      -0.994255912631277883d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.831584400419232317d0, &
      -0.831584400419232317d0, &
       0.831584400419232317d0, &
      -0.831584400419232317d0, &
       0.831584400419232317d0, &
      -0.831584400419232317d0, &
       0.831584400419232317d0, &
      -0.831584400419232317d0, &
       0.881813287779428800d0, &
      -0.881813287779428800d0, &
       0.881813287779428800d0, &
      -0.881813287779428800d0, &
       0.471598691151315974d0, &
      -0.471598691151315974d0, &
       0.471598691151315974d0, &
      -0.471598691151315974d0, &
       0.881813287779428800d0, &
      -0.881813287779428800d0, &
       0.881813287779428800d0, &
      -0.881813287779428800d0, &
       0.471598691151315974d0, &
      -0.471598691151315974d0, &
       0.471598691151315974d0, &
      -0.471598691151315974d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.977642811118264898d0, &
      -0.977642811118264898d0, &
       0.977642811118264898d0, &
      -0.977642811118264898d0, &
       0.210272522857306798d0, &
      -0.210272522857306798d0, &
       0.210272522857306798d0, &
      -0.210272522857306798d0, &
       0.977642811118264898d0, &
      -0.977642811118264898d0, &
       0.977642811118264898d0, &
      -0.977642811118264898d0, &
       0.210272522857306798d0, &
      -0.210272522857306798d0, &
       0.210272522857306798d0, &
      -0.210272522857306798d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.212646824707551862d0, &
       0.212646824707551862d0, &
      -0.212646824707551862d0, &
      -0.212646824707551862d0, &
       0.212646824707551862d0, &
       0.212646824707551862d0, &
      -0.212646824707551862d0, &
      -0.212646824707551862d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.967987158791472790d0, &
       0.967987158791472790d0, &
      -0.967987158791472790d0, &
      -0.967987158791472790d0, &
       0.967987158791472790d0, &
       0.967987158791472790d0, &
      -0.967987158791472790d0, &
      -0.967987158791472790d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.719016501040843470d0, &
       0.719016501040843470d0, &
      -0.719016501040843470d0, &
      -0.719016501040843470d0, &
       0.719016501040843470d0, &
       0.719016501040843470d0, &
      -0.719016501040843470d0, &
      -0.719016501040843470d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.407712664897769750d0, &
       0.407712664897769750d0, &
      -0.407712664897769750d0, &
      -0.407712664897769750d0, &
       0.407712664897769750d0, &
       0.407712664897769750d0, &
      -0.407712664897769750d0, &
      -0.407712664897769750d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.914472801120872480d0, &
       0.914472801120872480d0, &
      -0.914472801120872480d0, &
      -0.914472801120872480d0, &
       0.914472801120872480d0, &
       0.914472801120872480d0, &
      -0.914472801120872480d0, &
      -0.914472801120872480d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.994255912631277883d0, &
       0.994255912631277883d0, &
      -0.994255912631277883d0, &
      -0.994255912631277883d0, &
       0.994255912631277883d0, &
       0.994255912631277883d0, &
      -0.994255912631277883d0, &
      -0.994255912631277883d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.831584400419232317d0, &
       0.831584400419232317d0, &
      -0.831584400419232317d0, &
      -0.831584400419232317d0, &
       0.831584400419232317d0, &
       0.831584400419232317d0, &
      -0.831584400419232317d0, &
      -0.831584400419232317d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.471598691151315974d0, &
       0.471598691151315974d0, &
      -0.471598691151315974d0, &
      -0.471598691151315974d0, &
       0.881813287779428800d0, &
       0.881813287779428800d0, &
      -0.881813287779428800d0, &
      -0.881813287779428800d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.881813287779428800d0, &
      -0.881813287779428800d0, &
       0.881813287779428800d0, &
      -0.881813287779428800d0, &
       0.471598691151315974d0, &
      -0.471598691151315974d0, &
       0.471598691151315974d0, &
      -0.471598691151315974d0, &
       0.210272522857306798d0, &
       0.210272522857306798d0, &
      -0.210272522857306798d0, &
      -0.210272522857306798d0, &
       0.977642811118264898d0, &
       0.977642811118264898d0, &
      -0.977642811118264898d0, &
      -0.977642811118264898d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.977642811118264898d0, &
      -0.977642811118264898d0, &
       0.977642811118264898d0, &
      -0.977642811118264898d0, &
       0.210272522857306798d0, &
      -0.210272522857306798d0, &
       0.210272522857306798d0, &
      -0.210272522857306798d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.212646824707551862d0, &
       0.212646824707551862d0, &
       0.212646824707551862d0, &
       0.212646824707551862d0, &
      -0.212646824707551862d0, &
      -0.212646824707551862d0, &
      -0.212646824707551862d0, &
      -0.212646824707551862d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
       0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
      -0.690934630750911105d0, &
       0.967987158791472790d0, &
       0.967987158791472790d0, &
       0.967987158791472790d0, &
       0.967987158791472790d0, &
      -0.967987158791472790d0, &
      -0.967987158791472790d0, &
      -0.967987158791472790d0, &
      -0.967987158791472790d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
       0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
      -0.177483605460915794d0, &
       0.719016501040843470d0, &
       0.719016501040843470d0, &
       0.719016501040843470d0, &
       0.719016501040843470d0, &
      -0.719016501040843470d0, &
      -0.719016501040843470d0, &
      -0.719016501040843470d0, &
      -0.719016501040843470d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
       0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
      -0.491434263778474600d0, &
       0.407712664897769750d0, &
       0.407712664897769750d0, &
       0.407712664897769750d0, &
       0.407712664897769750d0, &
      -0.407712664897769750d0, &
      -0.407712664897769750d0, &
      -0.407712664897769750d0, &
      -0.407712664897769750d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
       0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
      -0.645666470742425558d0, &
       0.914472801120872480d0, &
       0.914472801120872480d0, &
       0.914472801120872480d0, &
       0.914472801120872480d0, &
      -0.914472801120872480d0, &
      -0.914472801120872480d0, &
      -0.914472801120872480d0, &
      -0.914472801120872480d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
       0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
      -0.286128901030763827d0, &
       0.994255912631277883d0, &
       0.994255912631277883d0, &
       0.994255912631277883d0, &
       0.994255912631277883d0, &
      -0.994255912631277883d0, &
      -0.994255912631277883d0, &
      -0.994255912631277883d0, &
      -0.994255912631277883d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
       0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
      -0.075680843671780185d0, &
       0.831584400419232317d0, &
       0.831584400419232317d0, &
       0.831584400419232317d0, &
       0.831584400419232317d0, &
      -0.831584400419232317d0, &
      -0.831584400419232317d0, &
      -0.831584400419232317d0, &
      -0.831584400419232317d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
       0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
      -0.392725976336800175d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.471598691151315974d0, &
       0.471598691151315974d0, &
      -0.471598691151315974d0, &
      -0.471598691151315974d0, &
       0.881813287779428800d0, &
       0.881813287779428800d0, &
      -0.881813287779428800d0, &
      -0.881813287779428800d0, &
       0.471598691151315974d0, &
       0.471598691151315974d0, &
      -0.471598691151315974d0, &
      -0.471598691151315974d0, &
       0.881813287779428800d0, &
       0.881813287779428800d0, &
      -0.881813287779428800d0, &
      -0.881813287779428800d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.210272522857306798d0, &
       0.210272522857306798d0, &
      -0.210272522857306798d0, &
      -0.210272522857306798d0, &
       0.977642811118264898d0, &
       0.977642811118264898d0, &
      -0.977642811118264898d0, &
      -0.977642811118264898d0, &
       0.210272522857306798d0, &
       0.210272522857306798d0, &
      -0.210272522857306798d0, &
      -0.210272522857306798d0, &
       0.977642811118264898d0, &
       0.977642811118264898d0, &
      -0.977642811118264898d0, &
      -0.977642811118264898d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
       0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
      -0.450233038258262552d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
       0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
      -0.868946032287241210d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
       0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
      -0.205482369640304391d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
       0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
      -0.106801826075804879d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
       0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
      -0.799927854385728554d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
       0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
      -0.590515704892527138d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
       0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
      -0.310428403516654350d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
       0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
      -0.771746262691590079d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
       0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
      -0.555015236107680665d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
       0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
      -0.099217696364292340d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
       0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0, &
      -0.334436314534345525d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
       0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0, &
      -0.937180985855372239d0/)
 
weight = (/&
       0.000526589796822444d0, &
       0.000526589796822444d0, &
       0.000526589796822444d0, &
       0.000526589796822444d0, &
       0.000526589796822444d0, &
       0.000526589796822444d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002548219972002607d0, &
       0.002512317418927307d0, &
       0.002512317418927307d0, &
       0.002512317418927307d0, &
       0.002512317418927307d0, &
       0.002512317418927307d0, &
       0.002512317418927307d0, &
       0.002512317418927307d0, &
       0.002512317418927307d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002530403801186355d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002014279020918528d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002501725168402936d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002513267174597564d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.002302694782227416d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.001462495621594614d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002445373437312980d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.002417442375638981d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.001910951282179532d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002416930044324775d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002512236854563495d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002496644054553086d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0, &
       0.002236607760437849d0/)

else if(integrationpoints == 350) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.024383301669355525d0, &
      -0.024383301669355525d0, &
       0.024383301669355525d0, &
      -0.024383301669355525d0, &
       0.024383301669355525d0, &
      -0.024383301669355525d0, &
       0.024383301669355525d0, &
      -0.024383301669355525d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.734996850587745554d0, &
      -0.734996850587745554d0, &
       0.734996850587745554d0, &
      -0.734996850587745554d0, &
       0.734996850587745554d0, &
      -0.734996850587745554d0, &
       0.734996850587745554d0, &
      -0.734996850587745554d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.962129055136014433d0, &
      -0.962129055136014433d0, &
       0.962129055136014433d0, &
      -0.962129055136014433d0, &
       0.962129055136014433d0, &
      -0.962129055136014433d0, &
       0.962129055136014433d0, &
      -0.962129055136014433d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.198501311223365423d0, &
      -0.198501311223365423d0, &
       0.198501311223365423d0, &
      -0.198501311223365423d0, &
       0.198501311223365423d0, &
      -0.198501311223365423d0, &
       0.198501311223365423d0, &
      -0.198501311223365423d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.860001812127547005d0, &
      -0.860001812127547005d0, &
       0.860001812127547005d0, &
      -0.860001812127547005d0, &
       0.860001812127547005d0, &
      -0.860001812127547005d0, &
       0.860001812127547005d0, &
      -0.860001812127547005d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.394199888605838877d0, &
      -0.394199888605838877d0, &
       0.394199888605838877d0, &
      -0.394199888605838877d0, &
       0.394199888605838877d0, &
      -0.394199888605838877d0, &
       0.394199888605838877d0, &
      -0.394199888605838877d0, &
       0.193294501323033913d0, &
      -0.193294501323033913d0, &
       0.193294501323033913d0, &
      -0.193294501323033913d0, &
       0.981140782843257231d0, &
      -0.981140782843257231d0, &
       0.981140782843257231d0, &
      -0.981140782843257231d0, &
       0.193294501323033913d0, &
      -0.193294501323033913d0, &
       0.193294501323033913d0, &
      -0.193294501323033913d0, &
       0.981140782843257231d0, &
      -0.981140782843257231d0, &
       0.981140782843257231d0, &
      -0.981140782843257231d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.380049491989930321d0, &
      -0.380049491989930321d0, &
       0.380049491989930321d0, &
      -0.380049491989930321d0, &
       0.924966152698678989d0, &
      -0.924966152698678989d0, &
       0.924966152698678989d0, &
      -0.924966152698678989d0, &
       0.380049491989930321d0, &
      -0.380049491989930321d0, &
       0.380049491989930321d0, &
      -0.380049491989930321d0, &
       0.924966152698678989d0, &
      -0.924966152698678989d0, &
       0.924966152698678989d0, &
      -0.924966152698678989d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.024383301669355525d0, &
       0.024383301669355525d0, &
      -0.024383301669355525d0, &
      -0.024383301669355525d0, &
       0.024383301669355525d0, &
       0.024383301669355525d0, &
      -0.024383301669355525d0, &
      -0.024383301669355525d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.734996850587745554d0, &
       0.734996850587745554d0, &
      -0.734996850587745554d0, &
      -0.734996850587745554d0, &
       0.734996850587745554d0, &
       0.734996850587745554d0, &
      -0.734996850587745554d0, &
      -0.734996850587745554d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.962129055136014433d0, &
       0.962129055136014433d0, &
      -0.962129055136014433d0, &
      -0.962129055136014433d0, &
       0.962129055136014433d0, &
       0.962129055136014433d0, &
      -0.962129055136014433d0, &
      -0.962129055136014433d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.198501311223365423d0, &
       0.198501311223365423d0, &
      -0.198501311223365423d0, &
      -0.198501311223365423d0, &
       0.198501311223365423d0, &
       0.198501311223365423d0, &
      -0.198501311223365423d0, &
      -0.198501311223365423d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.860001812127547005d0, &
       0.860001812127547005d0, &
      -0.860001812127547005d0, &
      -0.860001812127547005d0, &
       0.860001812127547005d0, &
       0.860001812127547005d0, &
      -0.860001812127547005d0, &
      -0.860001812127547005d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.394199888605838877d0, &
       0.394199888605838877d0, &
      -0.394199888605838877d0, &
      -0.394199888605838877d0, &
       0.394199888605838877d0, &
       0.394199888605838877d0, &
      -0.394199888605838877d0, &
      -0.394199888605838877d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.981140782843257231d0, &
       0.981140782843257231d0, &
      -0.981140782843257231d0, &
      -0.981140782843257231d0, &
       0.193294501323033913d0, &
       0.193294501323033913d0, &
      -0.193294501323033913d0, &
      -0.193294501323033913d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.193294501323033913d0, &
      -0.193294501323033913d0, &
       0.193294501323033913d0, &
      -0.193294501323033913d0, &
       0.981140782843257231d0, &
      -0.981140782843257231d0, &
       0.981140782843257231d0, &
      -0.981140782843257231d0, &
       0.924966152698678989d0, &
       0.924966152698678989d0, &
      -0.924966152698678989d0, &
      -0.924966152698678989d0, &
       0.380049491989930321d0, &
       0.380049491989930321d0, &
      -0.380049491989930321d0, &
      -0.380049491989930321d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.380049491989930321d0, &
      -0.380049491989930321d0, &
       0.380049491989930321d0, &
      -0.380049491989930321d0, &
       0.924966152698678989d0, &
      -0.924966152698678989d0, &
       0.924966152698678989d0, &
      -0.924966152698678989d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.024383301669355525d0, &
       0.024383301669355525d0, &
       0.024383301669355525d0, &
       0.024383301669355525d0, &
      -0.024383301669355525d0, &
      -0.024383301669355525d0, &
      -0.024383301669355525d0, &
      -0.024383301669355525d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
       0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
      -0.706896546391231562d0, &
       0.734996850587745554d0, &
       0.734996850587745554d0, &
       0.734996850587745554d0, &
       0.734996850587745554d0, &
      -0.734996850587745554d0, &
      -0.734996850587745554d0, &
      -0.734996850587745554d0, &
      -0.734996850587745554d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
       0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
      -0.479468262571202508d0, &
       0.962129055136014433d0, &
       0.962129055136014433d0, &
       0.962129055136014433d0, &
       0.962129055136014433d0, &
      -0.962129055136014433d0, &
      -0.962129055136014433d0, &
      -0.962129055136014433d0, &
      -0.962129055136014433d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
       0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
      -0.192753315487801891d0, &
       0.198501311223365423d0, &
       0.198501311223365423d0, &
       0.198501311223365423d0, &
       0.198501311223365423d0, &
      -0.198501311223365423d0, &
      -0.198501311223365423d0, &
      -0.198501311223365423d0, &
      -0.198501311223365423d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
       0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
      -0.693035796132712267d0, &
       0.860001812127547005d0, &
       0.860001812127547005d0, &
       0.860001812127547005d0, &
       0.860001812127547005d0, &
      -0.860001812127547005d0, &
      -0.860001812127547005d0, &
      -0.860001812127547005d0, &
      -0.860001812127547005d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
       0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
      -0.360830211552009117d0, &
       0.394199888605838877d0, &
       0.394199888605838877d0, &
       0.394199888605838877d0, &
       0.394199888605838877d0, &
      -0.394199888605838877d0, &
      -0.394199888605838877d0, &
      -0.394199888605838877d0, &
      -0.394199888605838877d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
       0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
      -0.649848616149616864d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.981140782843257231d0, &
       0.981140782843257231d0, &
      -0.981140782843257231d0, &
      -0.981140782843257231d0, &
       0.193294501323033913d0, &
       0.193294501323033913d0, &
      -0.193294501323033913d0, &
      -0.193294501323033913d0, &
       0.981140782843257231d0, &
       0.981140782843257231d0, &
      -0.981140782843257231d0, &
      -0.981140782843257231d0, &
       0.193294501323033913d0, &
       0.193294501323033913d0, &
      -0.193294501323033913d0, &
      -0.193294501323033913d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.924966152698678989d0, &
       0.924966152698678989d0, &
      -0.924966152698678989d0, &
      -0.924966152698678989d0, &
       0.380049491989930321d0, &
       0.380049491989930321d0, &
      -0.380049491989930321d0, &
      -0.380049491989930321d0, &
       0.924966152698678989d0, &
       0.924966152698678989d0, &
      -0.924966152698678989d0, &
      -0.924966152698678989d0, &
       0.380049491989930321d0, &
       0.380049491989930321d0, &
      -0.380049491989930321d0, &
      -0.380049491989930321d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
       0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
      -0.535123047718276124d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
       0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
      -0.793453785658231547d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
       0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
      -0.289955882549957411d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
       0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
      -0.552182074349399254d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
       0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
      -0.828080150668686232d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
       0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
      -0.096841214551039570d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
       0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
      -0.378009189874486784d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
       0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0, &
      -0.907465826530512665d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
       0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0, &
      -0.183343464704165904d0/)
 
weight = (/&
       0.003006796749453936d0, &
       0.003006796749453936d0, &
       0.003006796749453936d0, &
       0.003006796749453936d0, &
       0.003006796749453936d0, &
       0.003006796749453936d0, &
       0.003050627745650771d0, &
       0.003050627745650771d0, &
       0.003050627745650771d0, &
       0.003050627745650771d0, &
       0.003050627745650771d0, &
       0.003050627745650771d0, &
       0.003050627745650771d0, &
       0.003050627745650771d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.001621104600288991d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.003005701484901752d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002990992529653774d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002982170644107595d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.002721564237310992d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003033513795811141d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.003007949555218533d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002881964603055307d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.002958357626535696d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.003036020026407088d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0, &
       0.002832187403926303d0/)
 
else if(integrationpoints == 302) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.867643624544083392d0, &
      -0.867643624544083392d0, &
       0.867643624544083392d0, &
      -0.867643624544083392d0, &
       0.867643624544083392d0, &
      -0.867643624544083392d0, &
       0.867643624544083392d0, &
      -0.867643624544083392d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.371034178384820945d0, &
      -0.371034178384820945d0, &
       0.371034178384820945d0, &
      -0.371034178384820945d0, &
       0.371034178384820945d0, &
      -0.371034178384820945d0, &
       0.371034178384820945d0, &
      -0.371034178384820945d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.743452042987555739d0, &
      -0.743452042987555739d0, &
       0.743452042987555739d0, &
      -0.743452042987555739d0, &
       0.743452042987555739d0, &
      -0.743452042987555739d0, &
       0.743452042987555739d0, &
      -0.743452042987555739d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.990705621379408097d0, &
      -0.990705621379408097d0, &
       0.990705621379408097d0, &
      -0.990705621379408097d0, &
       0.990705621379408097d0, &
      -0.990705621379408097d0, &
       0.990705621379408097d0, &
      -0.990705621379408097d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.949454317226443134d0, &
      -0.949454317226443134d0, &
       0.949454317226443134d0, &
      -0.949454317226443134d0, &
       0.949454317226443134d0, &
      -0.949454317226443134d0, &
       0.949454317226443134d0, &
      -0.949454317226443134d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.129238672710514424d0, &
      -0.129238672710514424d0, &
       0.129238672710514424d0, &
      -0.129238672710514424d0, &
       0.129238672710514424d0, &
      -0.129238672710514424d0, &
       0.129238672710514424d0, &
      -0.129238672710514424d0, &
       0.264415288706066287d0, &
      -0.264415288706066287d0, &
       0.264415288706066287d0, &
      -0.264415288706066287d0, &
       0.964408914879205992d0, &
      -0.964408914879205992d0, &
       0.964408914879205992d0, &
      -0.964408914879205992d0, &
       0.264415288706066287d0, &
      -0.264415288706066287d0, &
       0.264415288706066287d0, &
      -0.264415288706066287d0, &
       0.964408914879205992d0, &
      -0.964408914879205992d0, &
       0.964408914879205992d0, &
      -0.964408914879205992d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.571895589187896070d0, &
      -0.571895589187896070d0, &
       0.571895589187896070d0, &
      -0.571895589187896070d0, &
       0.820326419827759334d0, &
      -0.820326419827759334d0, &
       0.820326419827759334d0, &
      -0.820326419827759334d0, &
       0.571895589187896070d0, &
      -0.571895589187896070d0, &
       0.571895589187896070d0, &
      -0.571895589187896070d0, &
       0.820326419827759334d0, &
      -0.820326419827759334d0, &
       0.820326419827759334d0, &
      -0.820326419827759334d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.867643624544083392d0, &
       0.867643624544083392d0, &
      -0.867643624544083392d0, &
      -0.867643624544083392d0, &
       0.867643624544083392d0, &
       0.867643624544083392d0, &
      -0.867643624544083392d0, &
      -0.867643624544083392d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.371034178384820945d0, &
       0.371034178384820945d0, &
      -0.371034178384820945d0, &
      -0.371034178384820945d0, &
       0.371034178384820945d0, &
       0.371034178384820945d0, &
      -0.371034178384820945d0, &
      -0.371034178384820945d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.743452042987555739d0, &
       0.743452042987555739d0, &
      -0.743452042987555739d0, &
      -0.743452042987555739d0, &
       0.743452042987555739d0, &
       0.743452042987555739d0, &
      -0.743452042987555739d0, &
      -0.743452042987555739d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.990705621379408097d0, &
       0.990705621379408097d0, &
      -0.990705621379408097d0, &
      -0.990705621379408097d0, &
       0.990705621379408097d0, &
       0.990705621379408097d0, &
      -0.990705621379408097d0, &
      -0.990705621379408097d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.949454317226443134d0, &
       0.949454317226443134d0, &
      -0.949454317226443134d0, &
      -0.949454317226443134d0, &
       0.949454317226443134d0, &
       0.949454317226443134d0, &
      -0.949454317226443134d0, &
      -0.949454317226443134d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.129238672710514424d0, &
       0.129238672710514424d0, &
      -0.129238672710514424d0, &
      -0.129238672710514424d0, &
       0.129238672710514424d0, &
       0.129238672710514424d0, &
      -0.129238672710514424d0, &
      -0.129238672710514424d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.964408914879205992d0, &
       0.964408914879205992d0, &
      -0.964408914879205992d0, &
      -0.964408914879205992d0, &
       0.264415288706066287d0, &
       0.264415288706066287d0, &
      -0.264415288706066287d0, &
      -0.264415288706066287d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.264415288706066287d0, &
      -0.264415288706066287d0, &
       0.264415288706066287d0, &
      -0.264415288706066287d0, &
       0.964408914879205992d0, &
      -0.964408914879205992d0, &
       0.964408914879205992d0, &
      -0.964408914879205992d0, &
       0.820326419827759334d0, &
       0.820326419827759334d0, &
      -0.820326419827759334d0, &
      -0.820326419827759334d0, &
       0.571895589187896070d0, &
       0.571895589187896070d0, &
      -0.571895589187896070d0, &
      -0.571895589187896070d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.571895589187896070d0, &
      -0.571895589187896070d0, &
       0.571895589187896070d0, &
      -0.571895589187896070d0, &
       0.820326419827759334d0, &
      -0.820326419827759334d0, &
       0.820326419827759334d0, &
      -0.820326419827759334d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.867643624544083392d0, &
       0.867643624544083392d0, &
       0.867643624544083392d0, &
       0.867643624544083392d0, &
      -0.867643624544083392d0, &
      -0.867643624544083392d0, &
      -0.867643624544083392d0, &
      -0.867643624544083392d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
       0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
      -0.351564034557010519d0, &
       0.371034178384820945d0, &
       0.371034178384820945d0, &
       0.371034178384820945d0, &
       0.371034178384820945d0, &
      -0.371034178384820945d0, &
      -0.371034178384820945d0, &
      -0.371034178384820945d0, &
      -0.371034178384820945d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
       0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
      -0.656632941021961236d0, &
       0.743452042987555739d0, &
       0.743452042987555739d0, &
       0.743452042987555739d0, &
       0.743452042987555739d0, &
      -0.743452042987555739d0, &
      -0.743452042987555739d0, &
      -0.743452042987555739d0, &
      -0.743452042987555739d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
       0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
      -0.472905413258100482d0, &
       0.990705621379408097d0, &
       0.990705621379408097d0, &
       0.990705621379408097d0, &
       0.990705621379408097d0, &
      -0.990705621379408097d0, &
      -0.990705621379408097d0, &
      -0.990705621379408097d0, &
      -0.990705621379408097d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
       0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
      -0.096183085226147838d0, &
       0.949454317226443134d0, &
       0.949454317226443134d0, &
       0.949454317226443134d0, &
       0.949454317226443134d0, &
      -0.949454317226443134d0, &
      -0.949454317226443134d0, &
      -0.949454317226443134d0, &
      -0.949454317226443134d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
       0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
      -0.221964523629417793d0, &
       0.129238672710514424d0, &
       0.129238672710514424d0, &
       0.129238672710514424d0, &
       0.129238672710514424d0, &
      -0.129238672710514424d0, &
      -0.129238672710514424d0, &
      -0.129238672710514424d0, &
      -0.129238672710514424d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
       0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
      -0.701176641608954543d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.964408914879205992d0, &
       0.964408914879205992d0, &
      -0.964408914879205992d0, &
      -0.964408914879205992d0, &
       0.264415288706066287d0, &
       0.264415288706066287d0, &
      -0.264415288706066287d0, &
      -0.264415288706066287d0, &
       0.964408914879205992d0, &
       0.964408914879205992d0, &
      -0.964408914879205992d0, &
      -0.964408914879205992d0, &
       0.264415288706066287d0, &
       0.264415288706066287d0, &
      -0.264415288706066287d0, &
      -0.264415288706066287d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.820326419827759334d0, &
       0.820326419827759334d0, &
      -0.820326419827759334d0, &
      -0.820326419827759334d0, &
       0.571895589187896070d0, &
       0.571895589187896070d0, &
      -0.571895589187896070d0, &
      -0.571895589187896070d0, &
       0.820326419827759334d0, &
       0.820326419827759334d0, &
      -0.820326419827759334d0, &
      -0.820326419827759334d0, &
       0.571895589187896070d0, &
       0.571895589187896070d0, &
      -0.571895589187896070d0, &
      -0.571895589187896070d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
       0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
      -0.544867737258077356d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
       0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
      -0.800072749407395145d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
       0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
      -0.251003475177046520d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
       0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
      -0.902442529533000415d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
       0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0, &
      -0.412772408316853079d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
       0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0, &
      -0.123354853258332703d0/)
 
weight = (/&
       0.000854591172512815d0, &
       0.000854591172512815d0, &
       0.000854591172512815d0, &
       0.000854591172512815d0, &
       0.000854591172512815d0, &
       0.000854591172512815d0, &
       0.003599119285025571d0, &
       0.003599119285025571d0, &
       0.003599119285025571d0, &
       0.003599119285025571d0, &
       0.003599119285025571d0, &
       0.003599119285025571d0, &
       0.003599119285025571d0, &
       0.003599119285025571d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003449788424305883d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003604822601419882d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.003576729661743367d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.002352101413689164d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003108953122413675d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.003650045807677255d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.002982344963171804d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003600820932216460d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003571540554273387d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0, &
       0.003392312205006170d0/)
 
else if(integrationpoints == 266) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.094575076403713099d0, &
      -0.094575076403713099d0, &
       0.094575076403713099d0, &
      -0.094575076403713099d0, &
       0.094575076403713099d0, &
      -0.094575076403713099d0, &
       0.094575076403713099d0, &
      -0.094575076403713099d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.989694807462905390d0, &
      -0.989694807462905390d0, &
       0.989694807462905390d0, &
      -0.989694807462905390d0, &
       0.989694807462905390d0, &
      -0.989694807462905390d0, &
       0.989694807462905390d0, &
      -0.989694807462905390d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.753673939250815716d0, &
      -0.753673939250815716d0, &
       0.753673939250815716d0, &
      -0.753673939250815716d0, &
       0.753673939250815716d0, &
      -0.753673939250815716d0, &
       0.753673939250815716d0, &
      -0.753673939250815716d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.886098344997499066d0, &
      -0.886098344997499066d0, &
       0.886098344997499066d0, &
      -0.886098344997499066d0, &
       0.886098344997499066d0, &
      -0.886098344997499066d0, &
       0.886098344997499066d0, &
      -0.886098344997499066d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.351315128564633428d0, &
      -0.351315128564633428d0, &
       0.351315128564633428d0, &
      -0.351315128564633428d0, &
       0.351315128564633428d0, &
      -0.351315128564633428d0, &
       0.351315128564633428d0, &
      -0.351315128564633428d0, &
       0.850650808352039878d0, &
      -0.850650808352039878d0, &
       0.850650808352039878d0, &
      -0.850650808352039878d0, &
       0.525731112119133703d0, &
      -0.525731112119133703d0, &
       0.525731112119133703d0, &
      -0.525731112119133703d0, &
       0.850650808352039878d0, &
      -0.850650808352039878d0, &
       0.850650808352039878d0, &
      -0.850650808352039878d0, &
       0.525731112119133703d0, &
      -0.525731112119133703d0, &
       0.525731112119133703d0, &
      -0.525731112119133703d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.094575076403713099d0, &
       0.094575076403713099d0, &
      -0.094575076403713099d0, &
      -0.094575076403713099d0, &
       0.094575076403713099d0, &
       0.094575076403713099d0, &
      -0.094575076403713099d0, &
      -0.094575076403713099d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.989694807462905390d0, &
       0.989694807462905390d0, &
      -0.989694807462905390d0, &
      -0.989694807462905390d0, &
       0.989694807462905390d0, &
       0.989694807462905390d0, &
      -0.989694807462905390d0, &
      -0.989694807462905390d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.753673939250815716d0, &
       0.753673939250815716d0, &
      -0.753673939250815716d0, &
      -0.753673939250815716d0, &
       0.753673939250815716d0, &
       0.753673939250815716d0, &
      -0.753673939250815716d0, &
      -0.753673939250815716d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.886098344997499066d0, &
       0.886098344997499066d0, &
      -0.886098344997499066d0, &
      -0.886098344997499066d0, &
       0.886098344997499066d0, &
       0.886098344997499066d0, &
      -0.886098344997499066d0, &
      -0.886098344997499066d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.351315128564633428d0, &
       0.351315128564633428d0, &
      -0.351315128564633428d0, &
      -0.351315128564633428d0, &
       0.351315128564633428d0, &
       0.351315128564633428d0, &
      -0.351315128564633428d0, &
      -0.351315128564633428d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.525731112119133703d0, &
       0.525731112119133703d0, &
      -0.525731112119133703d0, &
      -0.525731112119133703d0, &
       0.850650808352039878d0, &
       0.850650808352039878d0, &
      -0.850650808352039878d0, &
      -0.850650808352039878d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.850650808352039878d0, &
      -0.850650808352039878d0, &
       0.850650808352039878d0, &
      -0.850650808352039878d0, &
       0.525731112119133703d0, &
      -0.525731112119133703d0, &
       0.525731112119133703d0, &
      -0.525731112119133703d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.094575076403713099d0, &
       0.094575076403713099d0, &
       0.094575076403713099d0, &
       0.094575076403713099d0, &
      -0.094575076403713099d0, &
      -0.094575076403713099d0, &
      -0.094575076403713099d0, &
      -0.094575076403713099d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
       0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
      -0.703937339158547504d0, &
       0.989694807462905390d0, &
       0.989694807462905390d0, &
       0.989694807462905390d0, &
       0.989694807462905390d0, &
      -0.989694807462905390d0, &
      -0.989694807462905390d0, &
      -0.989694807462905390d0, &
      -0.989694807462905390d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
       0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
      -0.101252624857241402d0, &
       0.753673939250815716d0, &
       0.753673939250815716d0, &
       0.753673939250815716d0, &
       0.753673939250815716d0, &
      -0.753673939250815716d0, &
      -0.753673939250815716d0, &
      -0.753673939250815716d0, &
      -0.753673939250815716d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
       0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
      -0.464744872642053886d0, &
       0.886098344997499066d0, &
       0.886098344997499066d0, &
       0.886098344997499066d0, &
       0.886098344997499066d0, &
      -0.886098344997499066d0, &
      -0.886098344997499066d0, &
      -0.886098344997499066d0, &
      -0.886098344997499066d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
       0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
      -0.327742065497162893d0, &
       0.351315128564633428d0, &
       0.351315128564633428d0, &
       0.351315128564633428d0, &
       0.351315128564633428d0, &
      -0.351315128564633428d0, &
      -0.351315128564633428d0, &
      -0.351315128564633428d0, &
      -0.351315128564633428d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
       0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
      -0.662033866369997392d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.525731112119133703d0, &
       0.525731112119133703d0, &
      -0.525731112119133703d0, &
      -0.525731112119133703d0, &
       0.850650808352039878d0, &
       0.850650808352039878d0, &
      -0.850650808352039878d0, &
      -0.850650808352039878d0, &
       0.525731112119133703d0, &
       0.525731112119133703d0, &
      -0.525731112119133703d0, &
      -0.525731112119133703d0, &
       0.850650808352039878d0, &
       0.850650808352039878d0, &
      -0.850650808352039878d0, &
      -0.850650808352039878d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
       0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
      -0.939227929749915758d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
       0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
      -0.115311201100970095d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
       0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
      -0.323348454269289876d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
       0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
      -0.819343388819120277d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
       0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0, &
      -0.524493924092236496d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
       0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0, &
      -0.231479015871260102d0/)
 
weight = (/&
      -0.001313769127326952d0, &
      -0.001313769127326952d0, &
      -0.001313769127326952d0, &
      -0.001313769127326952d0, &
      -0.001313769127326952d0, &
      -0.001313769127326952d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
      -0.002522728704859336d0, &
       0.004186853881700583d0, &
       0.004186853881700583d0, &
       0.004186853881700583d0, &
       0.004186853881700583d0, &
       0.004186853881700583d0, &
       0.004186853881700583d0, &
       0.004186853881700583d0, &
       0.004186853881700583d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.005315167977810885d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004047142377086219d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.004112482394406990d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.003595584899758782d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004256131351428158d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004229582700647240d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004080914225780505d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0, &
       0.004071467593830964d0/)
 
else if(integrationpoints == 230) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.772289253148364052d0, &
      -0.772289253148364052d0, &
       0.772289253148364052d0, &
      -0.772289253148364052d0, &
       0.772289253148364052d0, &
      -0.772289253148364052d0, &
       0.772289253148364052d0, &
      -0.772289253148364052d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.934317778845811686d0, &
      -0.934317778845811686d0, &
       0.934317778845811686d0, &
      -0.934317778845811686d0, &
       0.934317778845811686d0, &
      -0.934317778845811686d0, &
       0.934317778845811686d0, &
      -0.934317778845811686d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.158302205463478313d0, &
      -0.158302205463478313d0, &
       0.158302205463478313d0, &
      -0.158302205463478313d0, &
       0.158302205463478313d0, &
      -0.158302205463478313d0, &
       0.158302205463478313d0, &
      -0.158302205463478313d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.363485684956727118d0, &
      -0.363485684956727118d0, &
       0.363485684956727118d0, &
      -0.363485684956727118d0, &
       0.363485684956727118d0, &
      -0.363485684956727118d0, &
       0.363485684956727118d0, &
      -0.363485684956727118d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.998367683967727459d0, &
      -0.998367683967727459d0, &
       0.998367683967727459d0, &
      -0.998367683967727459d0, &
       0.998367683967727459d0, &
      -0.998367683967727459d0, &
       0.998367683967727459d0, &
      -0.998367683967727459d0, &
       0.582384230971558448d0, &
      -0.582384230971558448d0, &
       0.582384230971558448d0, &
      -0.582384230971558448d0, &
       0.812913653173365280d0, &
      -0.812913653173365280d0, &
       0.812913653173365280d0, &
      -0.812913653173365280d0, &
       0.582384230971558448d0, &
      -0.582384230971558448d0, &
       0.582384230971558448d0, &
      -0.582384230971558448d0, &
       0.812913653173365280d0, &
      -0.812913653173365280d0, &
       0.812913653173365280d0, &
      -0.812913653173365280d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.354587739051868800d0, &
      -0.354587739051868800d0, &
       0.354587739051868800d0, &
      -0.354587739051868800d0, &
       0.935022745880592976d0, &
      -0.935022745880592976d0, &
       0.935022745880592976d0, &
      -0.935022745880592976d0, &
       0.354587739051868800d0, &
      -0.354587739051868800d0, &
       0.354587739051868800d0, &
      -0.354587739051868800d0, &
       0.935022745880592976d0, &
      -0.935022745880592976d0, &
       0.935022745880592976d0, &
      -0.935022745880592976d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.772289253148364052d0, &
       0.772289253148364052d0, &
      -0.772289253148364052d0, &
      -0.772289253148364052d0, &
       0.772289253148364052d0, &
       0.772289253148364052d0, &
      -0.772289253148364052d0, &
      -0.772289253148364052d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.934317778845811686d0, &
       0.934317778845811686d0, &
      -0.934317778845811686d0, &
      -0.934317778845811686d0, &
       0.934317778845811686d0, &
       0.934317778845811686d0, &
      -0.934317778845811686d0, &
      -0.934317778845811686d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.158302205463478313d0, &
       0.158302205463478313d0, &
      -0.158302205463478313d0, &
      -0.158302205463478313d0, &
       0.158302205463478313d0, &
       0.158302205463478313d0, &
      -0.158302205463478313d0, &
      -0.158302205463478313d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.363485684956727118d0, &
       0.363485684956727118d0, &
      -0.363485684956727118d0, &
      -0.363485684956727118d0, &
       0.363485684956727118d0, &
       0.363485684956727118d0, &
      -0.363485684956727118d0, &
      -0.363485684956727118d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.998367683967727459d0, &
       0.998367683967727459d0, &
      -0.998367683967727459d0, &
      -0.998367683967727459d0, &
       0.998367683967727459d0, &
       0.998367683967727459d0, &
      -0.998367683967727459d0, &
      -0.998367683967727459d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.812913653173365280d0, &
       0.812913653173365280d0, &
      -0.812913653173365280d0, &
      -0.812913653173365280d0, &
       0.582384230971558448d0, &
       0.582384230971558448d0, &
      -0.582384230971558448d0, &
      -0.582384230971558448d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.582384230971558448d0, &
      -0.582384230971558448d0, &
       0.582384230971558448d0, &
      -0.582384230971558448d0, &
       0.812913653173365280d0, &
      -0.812913653173365280d0, &
       0.812913653173365280d0, &
      -0.812913653173365280d0, &
       0.935022745880592976d0, &
       0.935022745880592976d0, &
      -0.935022745880592976d0, &
      -0.935022745880592976d0, &
       0.354587739051868800d0, &
       0.354587739051868800d0, &
      -0.354587739051868800d0, &
      -0.354587739051868800d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.354587739051868800d0, &
      -0.354587739051868800d0, &
       0.354587739051868800d0, &
      -0.354587739051868800d0, &
       0.935022745880592976d0, &
      -0.935022745880592976d0, &
       0.935022745880592976d0, &
      -0.935022745880592976d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.772289253148364052d0, &
       0.772289253148364052d0, &
       0.772289253148364052d0, &
       0.772289253148364052d0, &
      -0.772289253148364052d0, &
      -0.772289253148364052d0, &
      -0.772289253148364052d0, &
      -0.772289253148364052d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
       0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
      -0.449204468739761109d0, &
       0.934317778845811686d0, &
       0.934317778845811686d0, &
       0.934317778845811686d0, &
       0.934317778845811686d0, &
      -0.934317778845811686d0, &
      -0.934317778845811686d0, &
      -0.934317778845811686d0, &
      -0.934317778845811686d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
       0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
      -0.252041949021020117d0, &
       0.158302205463478313d0, &
       0.158302205463478313d0, &
       0.158302205463478313d0, &
       0.158302205463478313d0, &
      -0.158302205463478313d0, &
      -0.158302205463478313d0, &
      -0.158302205463478313d0, &
      -0.158302205463478313d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
       0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
      -0.698190665844724223d0, &
       0.363485684956727118d0, &
       0.363485684956727118d0, &
       0.363485684956727118d0, &
       0.363485684956727118d0, &
      -0.363485684956727118d0, &
      -0.363485684956727118d0, &
      -0.363485684956727118d0, &
      -0.363485684956727118d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
       0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
      -0.658740524346095979d0, &
       0.998367683967727459d0, &
       0.998367683967727459d0, &
       0.998367683967727459d0, &
       0.998367683967727459d0, &
      -0.998367683967727459d0, &
      -0.998367683967727459d0, &
      -0.998367683967727459d0, &
      -0.998367683967727459d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
       0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
      -0.040385440500976602d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.812913653173365280d0, &
       0.812913653173365280d0, &
      -0.812913653173365280d0, &
      -0.812913653173365280d0, &
       0.582384230971558448d0, &
       0.582384230971558448d0, &
      -0.582384230971558448d0, &
      -0.582384230971558448d0, &
       0.812913653173365280d0, &
       0.812913653173365280d0, &
      -0.812913653173365280d0, &
      -0.812913653173365280d0, &
       0.582384230971558448d0, &
       0.582384230971558448d0, &
      -0.582384230971558448d0, &
      -0.582384230971558448d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.935022745880592976d0, &
       0.935022745880592976d0, &
      -0.935022745880592976d0, &
      -0.935022745880592976d0, &
       0.354587739051868800d0, &
       0.354587739051868800d0, &
      -0.354587739051868800d0, &
      -0.354587739051868800d0, &
       0.935022745880592976d0, &
       0.935022745880592976d0, &
      -0.935022745880592976d0, &
      -0.935022745880592976d0, &
       0.354587739051868800d0, &
       0.354587739051868800d0, &
      -0.354587739051868800d0, &
      -0.354587739051868800d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
       0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
      -0.843636521068894307d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
       0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0, &
      -0.486466153588664685d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
       0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0, &
      -0.227218180899818711d0/)
 
weight = (/&
      -0.055226399197273249d0, &
      -0.055226399197273249d0, &
      -0.055226399197273249d0, &
      -0.055226399197273249d0, &
      -0.055226399197273249d0, &
      -0.055226399197273249d0, &
       0.004450274607445226d0, &
       0.004450274607445226d0, &
       0.004450274607445226d0, &
       0.004450274607445226d0, &
       0.004450274607445226d0, &
       0.004450274607445226d0, &
       0.004450274607445226d0, &
       0.004450274607445226d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.004496841067921404d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.005049153450478750d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.003976408018051883d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.004401400650381014d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.017245443505444009d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.004231083095357343d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.005198069864064399d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0, &
       0.004695720972568883d0/)
 
else if(integrationpoints == 194) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.314196994182586287d0, &
      -0.314196994182586287d0, &
       0.314196994182586287d0, &
      -0.314196994182586287d0, &
       0.314196994182586287d0, &
      -0.314196994182586287d0, &
       0.314196994182586287d0, &
      -0.314196994182586287d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.912509096867473724d0, &
      -0.912509096867473724d0, &
       0.912509096867473724d0, &
      -0.912509096867473724d0, &
       0.912509096867473724d0, &
      -0.912509096867473724d0, &
       0.912509096867473724d0, &
      -0.912509096867473724d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.777493219314767114d0, &
      -0.777493219314767114d0, &
       0.777493219314767114d0, &
      -0.777493219314767114d0, &
       0.777493219314767114d0, &
      -0.777493219314767114d0, &
       0.777493219314767114d0, &
      -0.777493219314767114d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.982972302707253220d0, &
      -0.982972302707253220d0, &
       0.982972302707253220d0, &
      -0.982972302707253220d0, &
       0.982972302707253220d0, &
      -0.982972302707253220d0, &
       0.982972302707253220d0, &
      -0.982972302707253220d0, &
       0.345770219761128317d0, &
      -0.345770219761128317d0, &
       0.345770219761128317d0, &
      -0.345770219761128317d0, &
       0.938319218137591560d0, &
      -0.938319218137591560d0, &
       0.938319218137591560d0, &
      -0.938319218137591560d0, &
       0.345770219761128317d0, &
      -0.345770219761128317d0, &
       0.345770219761128317d0, &
      -0.345770219761128317d0, &
       0.938319218137591560d0, &
      -0.938319218137591560d0, &
       0.938319218137591560d0, &
      -0.938319218137591560d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.314196994182586287d0, &
       0.314196994182586287d0, &
      -0.314196994182586287d0, &
      -0.314196994182586287d0, &
       0.314196994182586287d0, &
       0.314196994182586287d0, &
      -0.314196994182586287d0, &
      -0.314196994182586287d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.912509096867473724d0, &
       0.912509096867473724d0, &
      -0.912509096867473724d0, &
      -0.912509096867473724d0, &
       0.912509096867473724d0, &
       0.912509096867473724d0, &
      -0.912509096867473724d0, &
      -0.912509096867473724d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.777493219314767114d0, &
       0.777493219314767114d0, &
      -0.777493219314767114d0, &
      -0.777493219314767114d0, &
       0.777493219314767114d0, &
       0.777493219314767114d0, &
      -0.777493219314767114d0, &
      -0.777493219314767114d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.982972302707253220d0, &
       0.982972302707253220d0, &
      -0.982972302707253220d0, &
      -0.982972302707253220d0, &
       0.982972302707253220d0, &
       0.982972302707253220d0, &
      -0.982972302707253220d0, &
      -0.982972302707253220d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.938319218137591560d0, &
       0.938319218137591560d0, &
      -0.938319218137591560d0, &
      -0.938319218137591560d0, &
       0.345770219761128317d0, &
       0.345770219761128317d0, &
      -0.345770219761128317d0, &
      -0.345770219761128317d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.345770219761128317d0, &
      -0.345770219761128317d0, &
       0.345770219761128317d0, &
      -0.345770219761128317d0, &
       0.938319218137591560d0, &
      -0.938319218137591560d0, &
       0.938319218137591560d0, &
      -0.938319218137591560d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.314196994182586287d0, &
       0.314196994182586287d0, &
       0.314196994182586287d0, &
       0.314196994182586287d0, &
      -0.314196994182586287d0, &
      -0.314196994182586287d0, &
      -0.314196994182586287d0, &
      -0.314196994182586287d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
       0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
      -0.671297344269522589d0, &
       0.912509096867473724d0, &
       0.912509096867473724d0, &
       0.912509096867473724d0, &
       0.912509096867473724d0, &
      -0.912509096867473724d0, &
      -0.912509096867473724d0, &
      -0.912509096867473724d0, &
      -0.912509096867473724d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
       0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
      -0.289246562757543901d0, &
       0.777493219314767114d0, &
       0.777493219314767114d0, &
       0.777493219314767114d0, &
       0.777493219314767114d0, &
      -0.777493219314767114d0, &
      -0.777493219314767114d0, &
      -0.777493219314767114d0, &
      -0.777493219314767114d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
       0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
      -0.444693317871743710d0, &
       0.982972302707253220d0, &
       0.982972302707253220d0, &
       0.982972302707253220d0, &
       0.982972302707253220d0, &
      -0.982972302707253220d0, &
      -0.982972302707253220d0, &
      -0.982972302707253220d0, &
      -0.982972302707253220d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
       0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
      -0.129933544765006709d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.938319218137591560d0, &
       0.938319218137591560d0, &
      -0.938319218137591560d0, &
      -0.938319218137591560d0, &
       0.345770219761128317d0, &
       0.345770219761128317d0, &
      -0.345770219761128317d0, &
      -0.345770219761128317d0, &
       0.938319218137591560d0, &
       0.938319218137591560d0, &
      -0.938319218137591560d0, &
      -0.938319218137591560d0, &
       0.345770219761128317d0, &
       0.345770219761128317d0, &
      -0.345770219761128317d0, &
      -0.345770219761128317d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
       0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
      -0.525118572443642018d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
       0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0, &
      -0.836036015482458872d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
       0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0, &
      -0.159041710538353004d0/)
 
weight = (/&
       0.001782340447244611d0, &
       0.001782340447244611d0, &
       0.001782340447244611d0, &
       0.001782340447244611d0, &
       0.001782340447244611d0, &
       0.001782340447244611d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005716905949977102d0, &
       0.005573383178848738d0, &
       0.005573383178848738d0, &
       0.005573383178848738d0, &
       0.005573383178848738d0, &
       0.005573383178848738d0, &
       0.005573383178848738d0, &
       0.005573383178848738d0, &
       0.005573383178848738d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005608704082587997d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005158237711805383d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.005518771467273614d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.004106777028169394d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005051846064614808d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0, &
       0.005530248916233094d0/)
 
else if(integrationpoints == 170) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.932642590312690589d0, &
      -0.932642590312690589d0, &
       0.932642590312690589d0, &
      -0.932642590312690589d0, &
       0.932642590312690589d0, &
      -0.932642590312690589d0, &
       0.932642590312690589d0, &
      -0.932642590312690589d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.300793595137701486d0, &
      -0.300793595137701486d0, &
       0.300793595137701486d0, &
      -0.300793595137701486d0, &
       0.300793595137701486d0, &
      -0.300793595137701486d0, &
       0.300793595137701486d0, &
      -0.300793595137701486d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.791795559393492110d0, &
      -0.791795559393492110d0, &
       0.791795559393492110d0, &
      -0.791795559393492110d0, &
       0.791795559393492110d0, &
      -0.791795559393492110d0, &
       0.791795559393492110d0, &
      -0.791795559393492110d0, &
       0.261393136033598816d0, &
      -0.261393136033598816d0, &
       0.261393136033598816d0, &
      -0.261393136033598816d0, &
       0.965232421976448429d0, &
      -0.965232421976448429d0, &
       0.965232421976448429d0, &
      -0.965232421976448429d0, &
       0.261393136033598816d0, &
      -0.261393136033598816d0, &
       0.261393136033598816d0, &
      -0.261393136033598816d0, &
       0.965232421976448429d0, &
      -0.965232421976448429d0, &
       0.965232421976448429d0, &
      -0.965232421976448429d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.932642590312690589d0, &
       0.932642590312690589d0, &
      -0.932642590312690589d0, &
      -0.932642590312690589d0, &
       0.932642590312690589d0, &
       0.932642590312690589d0, &
      -0.932642590312690589d0, &
      -0.932642590312690589d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.300793595137701486d0, &
       0.300793595137701486d0, &
      -0.300793595137701486d0, &
      -0.300793595137701486d0, &
       0.300793595137701486d0, &
       0.300793595137701486d0, &
      -0.300793595137701486d0, &
      -0.300793595137701486d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.791795559393492110d0, &
       0.791795559393492110d0, &
      -0.791795559393492110d0, &
      -0.791795559393492110d0, &
       0.791795559393492110d0, &
       0.791795559393492110d0, &
      -0.791795559393492110d0, &
      -0.791795559393492110d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.965232421976448429d0, &
       0.965232421976448429d0, &
      -0.965232421976448429d0, &
      -0.965232421976448429d0, &
       0.261393136033598816d0, &
       0.261393136033598816d0, &
      -0.261393136033598816d0, &
      -0.261393136033598816d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.261393136033598816d0, &
      -0.261393136033598816d0, &
       0.261393136033598816d0, &
      -0.261393136033598816d0, &
       0.965232421976448429d0, &
      -0.965232421976448429d0, &
       0.965232421976448429d0, &
      -0.965232421976448429d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.932642590312690589d0, &
       0.932642590312690589d0, &
       0.932642590312690589d0, &
       0.932642590312690589d0, &
      -0.932642590312690589d0, &
      -0.932642590312690589d0, &
      -0.932642590312690589d0, &
      -0.932642590312690589d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
       0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
      -0.255125262111413398d0, &
       0.300793595137701486d0, &
       0.300793595137701486d0, &
       0.300793595137701486d0, &
       0.300793595137701486d0, &
      -0.300793595137701486d0, &
      -0.300793595137701486d0, &
      -0.300793595137701486d0, &
      -0.300793595137701486d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
       0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
      -0.674360146036276586d0, &
       0.791795559393492110d0, &
       0.791795559393492110d0, &
       0.791795559393492110d0, &
       0.791795559393492110d0, &
      -0.791795559393492110d0, &
      -0.791795559393492110d0, &
      -0.791795559393492110d0, &
      -0.791795559393492110d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
       0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
      -0.431891069671941019d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.965232421976448429d0, &
       0.965232421976448429d0, &
      -0.965232421976448429d0, &
      -0.965232421976448429d0, &
       0.261393136033598816d0, &
       0.261393136033598816d0, &
      -0.261393136033598816d0, &
      -0.261393136033598816d0, &
       0.965232421976448429d0, &
       0.965232421976448429d0, &
      -0.965232421976448429d0, &
      -0.965232421976448429d0, &
       0.261393136033598816d0, &
       0.261393136033598816d0, &
      -0.261393136033598816d0, &
      -0.261393136033598816d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
       0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
      -0.854415804684658831d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
       0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0, &
      -0.144663074432511501d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
       0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0, &
      -0.499045316179603682d0/)
 
weight = (/&
       0.005544842902037365d0, &
       0.005544842902037365d0, &
       0.005544842902037365d0, &
       0.005544842902037365d0, &
       0.005544842902037365d0, &
       0.005544842902037365d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006071332770670752d0, &
       0.006383674773515093d0, &
       0.006383674773515093d0, &
       0.006383674773515093d0, &
       0.006383674773515093d0, &
       0.006383674773515093d0, &
       0.006383674773515093d0, &
       0.006383674773515093d0, &
       0.006383674773515093d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.005183387587747790d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006317929009813725d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.006201670006589077d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005477143385137348d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0, &
       0.005968383987681156d0/)
 
else if(integrationpoints == 146) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.291298882209526822d0, &
      -0.291298882209526822d0, &
       0.291298882209526822d0, &
      -0.291298882209526822d0, &
       0.291298882209526822d0, &
      -0.291298882209526822d0, &
       0.291298882209526822d0, &
      -0.291298882209526822d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.807089818359582600d0, &
      -0.807089818359582600d0, &
       0.807089818359582600d0, &
      -0.807089818359582600d0, &
       0.807089818359582600d0, &
      -0.807089818359582600d0, &
       0.807089818359582600d0, &
      -0.807089818359582600d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.974888643677173206d0, &
      -0.974888643677173206d0, &
       0.974888643677173206d0, &
      -0.974888643677173206d0, &
       0.974888643677173206d0, &
      -0.974888643677173206d0, &
       0.974888643677173206d0, &
      -0.974888643677173206d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.291298882209526822d0, &
       0.291298882209526822d0, &
      -0.291298882209526822d0, &
      -0.291298882209526822d0, &
       0.291298882209526822d0, &
       0.291298882209526822d0, &
      -0.291298882209526822d0, &
      -0.291298882209526822d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.807089818359582600d0, &
       0.807089818359582600d0, &
      -0.807089818359582600d0, &
      -0.807089818359582600d0, &
       0.807089818359582600d0, &
       0.807089818359582600d0, &
      -0.807089818359582600d0, &
      -0.807089818359582600d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.974888643677173206d0, &
       0.974888643677173206d0, &
      -0.974888643677173206d0, &
      -0.974888643677173206d0, &
       0.974888643677173206d0, &
       0.974888643677173206d0, &
      -0.974888643677173206d0, &
      -0.974888643677173206d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.291298882209526822d0, &
       0.291298882209526822d0, &
       0.291298882209526822d0, &
       0.291298882209526822d0, &
      -0.291298882209526822d0, &
      -0.291298882209526822d0, &
      -0.291298882209526822d0, &
      -0.291298882209526822d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
       0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
      -0.676441040011426353d0, &
       0.807089818359582600d0, &
       0.807089818359582600d0, &
       0.807089818359582600d0, &
       0.807089818359582600d0, &
      -0.807089818359582600d0, &
      -0.807089818359582600d0, &
      -0.807089818359582600d0, &
      -0.807089818359582600d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
       0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
      -0.417496122796545299d0, &
       0.974888643677173206d0, &
       0.974888643677173206d0, &
       0.974888643677173206d0, &
       0.974888643677173206d0, &
      -0.974888643677173206d0, &
      -0.974888643677173206d0, &
      -0.974888643677173206d0, &
      -0.974888643677173206d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
       0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
      -0.157467667203908201d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
       0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
      -0.882270011260322673d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
       0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0, &
      -0.449332832326955700d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
       0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0, &
      -0.140355381171318294d0/)
 
weight = (/&
       0.000599631368862138d0, &
       0.000599631368862138d0, &
       0.000599631368862138d0, &
       0.000599631368862138d0, &
       0.000599631368862138d0, &
       0.000599631368862138d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007372999718620756d0, &
       0.007210515360144488d0, &
       0.007210515360144488d0, &
       0.007210515360144488d0, &
       0.007210515360144488d0, &
       0.007210515360144488d0, &
       0.007210515360144488d0, &
       0.007210515360144488d0, &
       0.007210515360144488d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.007116355493117555d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.006753829486314477d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.007574394159054034d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0, &
       0.006991087353303262d0/)
 
else if(integrationpoints == 110) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.965124035086594056d0, &
      -0.965124035086594056d0, &
       0.965124035086594056d0, &
      -0.965124035086594056d0, &
       0.965124035086594056d0, &
      -0.965124035086594056d0, &
       0.965124035086594056d0, &
      -0.965124035086594056d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.215957291845848443d0, &
      -0.215957291845848443d0, &
       0.215957291845848443d0, &
      -0.215957291845848443d0, &
       0.215957291845848443d0, &
      -0.215957291845848443d0, &
       0.215957291845848443d0, &
      -0.215957291845848443d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.828769981252592269d0, &
      -0.828769981252592269d0, &
       0.828769981252592269d0, &
      -0.828769981252592269d0, &
       0.828769981252592269d0, &
      -0.828769981252592269d0, &
       0.828769981252592269d0, &
      -0.828769981252592269d0, &
       0.478369028812150210d0, &
      -0.478369028812150210d0, &
       0.478369028812150210d0, &
      -0.478369028812150210d0, &
       0.878158910604066145d0, &
      -0.878158910604066145d0, &
       0.878158910604066145d0, &
      -0.878158910604066145d0, &
       0.478369028812150210d0, &
      -0.478369028812150210d0, &
       0.478369028812150210d0, &
      -0.478369028812150210d0, &
       0.878158910604066145d0, &
      -0.878158910604066145d0, &
       0.878158910604066145d0, &
      -0.878158910604066145d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.965124035086594056d0, &
       0.965124035086594056d0, &
      -0.965124035086594056d0, &
      -0.965124035086594056d0, &
       0.965124035086594056d0, &
       0.965124035086594056d0, &
      -0.965124035086594056d0, &
      -0.965124035086594056d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.215957291845848443d0, &
       0.215957291845848443d0, &
      -0.215957291845848443d0, &
      -0.215957291845848443d0, &
       0.215957291845848443d0, &
       0.215957291845848443d0, &
      -0.215957291845848443d0, &
      -0.215957291845848443d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.828769981252592269d0, &
       0.828769981252592269d0, &
      -0.828769981252592269d0, &
      -0.828769981252592269d0, &
       0.828769981252592269d0, &
       0.828769981252592269d0, &
      -0.828769981252592269d0, &
      -0.828769981252592269d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.878158910604066145d0, &
       0.878158910604066145d0, &
      -0.878158910604066145d0, &
      -0.878158910604066145d0, &
       0.478369028812150210d0, &
       0.478369028812150210d0, &
      -0.478369028812150210d0, &
      -0.478369028812150210d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.478369028812150210d0, &
      -0.478369028812150210d0, &
       0.478369028812150210d0, &
      -0.478369028812150210d0, &
       0.878158910604066145d0, &
      -0.878158910604066145d0, &
       0.878158910604066145d0, &
      -0.878158910604066145d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.965124035086594056d0, &
       0.965124035086594056d0, &
       0.965124035086594056d0, &
       0.965124035086594056d0, &
      -0.965124035086594056d0, &
      -0.965124035086594056d0, &
      -0.965124035086594056d0, &
      -0.965124035086594056d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
       0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
      -0.185115635344736212d0, &
       0.215957291845848443d0, &
       0.215957291845848443d0, &
       0.215957291845848443d0, &
       0.215957291845848443d0, &
      -0.215957291845848443d0, &
      -0.215957291845848443d0, &
      -0.215957291845848443d0, &
      -0.215957291845848443d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
       0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
      -0.690421048382292235d0, &
       0.828769981252592269d0, &
       0.828769981252592269d0, &
       0.828769981252592269d0, &
       0.828769981252592269d0, &
      -0.828769981252592269d0, &
      -0.828769981252592269d0, &
      -0.828769981252592269d0, &
      -0.828769981252592269d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
       0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
      -0.395689473055941876d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.878158910604066145d0, &
       0.878158910604066145d0, &
      -0.878158910604066145d0, &
      -0.878158910604066145d0, &
       0.478369028812150210d0, &
       0.478369028812150210d0, &
      -0.478369028812150210d0, &
      -0.478369028812150210d0, &
       0.878158910604066145d0, &
       0.878158910604066145d0, &
      -0.878158910604066145d0, &
      -0.878158910604066145d0, &
       0.478369028812150210d0, &
       0.478369028812150210d0, &
      -0.478369028812150210d0, &
      -0.478369028812150210d0/)
 
weight = (/&
       0.003828270494937162d0, &
       0.003828270494937162d0, &
       0.003828270494937162d0, &
       0.003828270494937162d0, &
       0.003828270494937162d0, &
       0.003828270494937162d0, &
       0.009793737512487513d0, &
       0.009793737512487513d0, &
       0.009793737512487513d0, &
       0.009793737512487513d0, &
       0.009793737512487513d0, &
       0.009793737512487513d0, &
       0.009793737512487513d0, &
       0.009793737512487513d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.008211737283191111d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009942814891178103d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009595471336070962d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0, &
       0.009694996361663029d0/)
 
else if(integrationpoints ==  86) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.852518311701267573d0, &
      -0.852518311701267573d0, &
       0.852518311701267573d0, &
      -0.852518311701267573d0, &
       0.852518311701267573d0, &
      -0.852518311701267573d0, &
       0.852518311701267573d0, &
      -0.852518311701267573d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.189063552885394975d0, &
      -0.189063552885394975d0, &
       0.189063552885394975d0, &
      -0.189063552885394975d0, &
       0.189063552885394975d0, &
      -0.189063552885394975d0, &
       0.189063552885394975d0, &
      -0.189063552885394975d0, &
       0.374243039090341179d0, &
      -0.374243039090341179d0, &
       0.374243039090341179d0, &
      -0.374243039090341179d0, &
       0.927330657151172466d0, &
      -0.927330657151172466d0, &
       0.927330657151172466d0, &
      -0.927330657151172466d0, &
       0.374243039090341179d0, &
      -0.374243039090341179d0, &
       0.374243039090341179d0, &
      -0.374243039090341179d0, &
       0.927330657151172466d0, &
      -0.927330657151172466d0, &
       0.927330657151172466d0, &
      -0.927330657151172466d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.852518311701267573d0, &
       0.852518311701267573d0, &
      -0.852518311701267573d0, &
      -0.852518311701267573d0, &
       0.852518311701267573d0, &
       0.852518311701267573d0, &
      -0.852518311701267573d0, &
      -0.852518311701267573d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.189063552885394975d0, &
       0.189063552885394975d0, &
      -0.189063552885394975d0, &
      -0.189063552885394975d0, &
       0.189063552885394975d0, &
       0.189063552885394975d0, &
      -0.189063552885394975d0, &
      -0.189063552885394975d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.927330657151172466d0, &
       0.927330657151172466d0, &
      -0.927330657151172466d0, &
      -0.927330657151172466d0, &
       0.374243039090341179d0, &
       0.374243039090341179d0, &
      -0.374243039090341179d0, &
      -0.374243039090341179d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.374243039090341179d0, &
      -0.374243039090341179d0, &
       0.374243039090341179d0, &
      -0.374243039090341179d0, &
       0.927330657151172466d0, &
      -0.927330657151172466d0, &
       0.927330657151172466d0, &
      -0.927330657151172466d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.852518311701267573d0, &
       0.852518311701267573d0, &
       0.852518311701267573d0, &
       0.852518311701267573d0, &
      -0.852518311701267573d0, &
      -0.852518311701267573d0, &
      -0.852518311701267573d0, &
      -0.852518311701267573d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
       0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
      -0.369602846454150202d0, &
       0.189063552885394975d0, &
       0.189063552885394975d0, &
       0.189063552885394975d0, &
       0.189063552885394975d0, &
      -0.189063552885394975d0, &
      -0.189063552885394975d0, &
      -0.189063552885394975d0, &
      -0.189063552885394975d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
       0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
      -0.694354006602666440d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.927330657151172466d0, &
       0.927330657151172466d0, &
      -0.927330657151172466d0, &
      -0.927330657151172466d0, &
       0.374243039090341179d0, &
       0.374243039090341179d0, &
      -0.374243039090341179d0, &
      -0.374243039090341179d0, &
       0.927330657151172466d0, &
       0.927330657151172466d0, &
      -0.927330657151172466d0, &
      -0.927330657151172466d0, &
       0.374243039090341179d0, &
       0.374243039090341179d0, &
      -0.374243039090341179d0, &
      -0.374243039090341179d0/)
 
weight = (/&
       0.011544011544011541d0, &
       0.011544011544011541d0, &
       0.011544011544011541d0, &
       0.011544011544011541d0, &
       0.011544011544011541d0, &
       0.011544011544011541d0, &
       0.011943909085856280d0, &
       0.011943909085856280d0, &
       0.011943909085856280d0, &
       0.011943909085856280d0, &
       0.011943909085856280d0, &
       0.011943909085856280d0, &
       0.011943909085856280d0, &
       0.011943909085856280d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011110555710603400d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011876501294537141d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0, &
       0.011812303746904481d0/)
 
else if(integrationpoints ==  74) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.733799385705342755d0, &
      -0.733799385705342755d0, &
       0.733799385705342755d0, &
      -0.733799385705342755d0, &
       0.733799385705342755d0, &
      -0.733799385705342755d0, &
       0.733799385705342755d0, &
      -0.733799385705342755d0, &
       0.320772648980776398d0, &
      -0.320772648980776398d0, &
       0.320772648980776398d0, &
      -0.320772648980776398d0, &
       0.947156221362587925d0, &
      -0.947156221362587925d0, &
       0.947156221362587925d0, &
      -0.947156221362587925d0, &
       0.320772648980776398d0, &
      -0.320772648980776398d0, &
       0.320772648980776398d0, &
      -0.320772648980776398d0, &
       0.947156221362587925d0, &
      -0.947156221362587925d0, &
       0.947156221362587925d0, &
      -0.947156221362587925d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.733799385705342755d0, &
       0.733799385705342755d0, &
      -0.733799385705342755d0, &
      -0.733799385705342755d0, &
       0.733799385705342755d0, &
       0.733799385705342755d0, &
      -0.733799385705342755d0, &
      -0.733799385705342755d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.947156221362587925d0, &
       0.947156221362587925d0, &
      -0.947156221362587925d0, &
      -0.947156221362587925d0, &
       0.320772648980776398d0, &
       0.320772648980776398d0, &
      -0.320772648980776398d0, &
      -0.320772648980776398d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.320772648980776398d0, &
      -0.320772648980776398d0, &
       0.320772648980776398d0, &
      -0.320772648980776398d0, &
       0.947156221362587925d0, &
      -0.947156221362587925d0, &
       0.947156221362587925d0, &
      -0.947156221362587925d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.733799385705342755d0, &
       0.733799385705342755d0, &
       0.733799385705342755d0, &
       0.733799385705342755d0, &
      -0.733799385705342755d0, &
      -0.733799385705342755d0, &
      -0.733799385705342755d0, &
      -0.733799385705342755d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
       0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
      -0.480384461415261421d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.947156221362587925d0, &
       0.947156221362587925d0, &
      -0.947156221362587925d0, &
      -0.947156221362587925d0, &
       0.320772648980776398d0, &
       0.320772648980776398d0, &
      -0.320772648980776398d0, &
      -0.320772648980776398d0, &
       0.947156221362587925d0, &
       0.947156221362587925d0, &
      -0.947156221362587925d0, &
      -0.947156221362587925d0, &
       0.320772648980776398d0, &
       0.320772648980776398d0, &
      -0.320772648980776398d0, &
      -0.320772648980776398d0/)
 
weight = (/&
       0.000513067179733846d0, &
       0.000513067179733846d0, &
       0.000513067179733846d0, &
       0.000513067179733846d0, &
       0.000513067179733846d0, &
       0.000513067179733846d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
       0.016604069565742039d0, &
      -0.029586038961038959d0, &
      -0.029586038961038959d0, &
      -0.029586038961038959d0, &
      -0.029586038961038959d0, &
      -0.029586038961038959d0, &
      -0.029586038961038959d0, &
      -0.029586038961038959d0, &
      -0.029586038961038959d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.026576207082159461d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0, &
       0.016522170993715710d0/)
 
else if(integrationpoints ==  50) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.904534033733290888d0, &
      -0.904534033733290888d0, &
       0.904534033733290888d0, &
      -0.904534033733290888d0, &
       0.904534033733290888d0, &
      -0.904534033733290888d0, &
       0.904534033733290888d0, &
      -0.904534033733290888d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.904534033733290888d0, &
       0.904534033733290888d0, &
      -0.904534033733290888d0, &
      -0.904534033733290888d0, &
       0.904534033733290888d0, &
       0.904534033733290888d0, &
      -0.904534033733290888d0, &
      -0.904534033733290888d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.904534033733290888d0, &
       0.904534033733290888d0, &
       0.904534033733290888d0, &
       0.904534033733290888d0, &
      -0.904534033733290888d0, &
      -0.904534033733290888d0, &
      -0.904534033733290888d0, &
      -0.904534033733290888d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
       0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0, &
      -0.301511344577763574d0/)
 
weight = (/&
       0.012698412698412700d0, &
       0.012698412698412700d0, &
       0.012698412698412700d0, &
       0.012698412698412700d0, &
       0.012698412698412700d0, &
       0.012698412698412700d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.022574955908289240d0, &
       0.021093750000000001d0, &
       0.021093750000000001d0, &
       0.021093750000000001d0, &
       0.021093750000000001d0, &
       0.021093750000000001d0, &
       0.021093750000000001d0, &
       0.021093750000000001d0, &
       0.021093750000000001d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0, &
       0.020173335537918870d0/)
 
else if(integrationpoints ==  38) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.459700843380983104d0, &
      -0.459700843380983104d0, &
       0.459700843380983104d0, &
      -0.459700843380983104d0, &
       0.888073833977115257d0, &
      -0.888073833977115257d0, &
       0.888073833977115257d0, &
      -0.888073833977115257d0, &
       0.459700843380983104d0, &
      -0.459700843380983104d0, &
       0.459700843380983104d0, &
      -0.459700843380983104d0, &
       0.888073833977115257d0, &
      -0.888073833977115257d0, &
       0.888073833977115257d0, &
      -0.888073833977115257d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.888073833977115257d0, &
       0.888073833977115257d0, &
      -0.888073833977115257d0, &
      -0.888073833977115257d0, &
       0.459700843380983104d0, &
       0.459700843380983104d0, &
      -0.459700843380983104d0, &
      -0.459700843380983104d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.459700843380983104d0, &
      -0.459700843380983104d0, &
       0.459700843380983104d0, &
      -0.459700843380983104d0, &
       0.888073833977115257d0, &
      -0.888073833977115257d0, &
       0.888073833977115257d0, &
      -0.888073833977115257d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.888073833977115257d0, &
       0.888073833977115257d0, &
      -0.888073833977115257d0, &
      -0.888073833977115257d0, &
       0.459700843380983104d0, &
       0.459700843380983104d0, &
      -0.459700843380983104d0, &
      -0.459700843380983104d0, &
       0.888073833977115257d0, &
       0.888073833977115257d0, &
      -0.888073833977115257d0, &
      -0.888073833977115257d0, &
       0.459700843380983104d0, &
       0.459700843380983104d0, &
      -0.459700843380983104d0, &
      -0.459700843380983104d0/)
 
weight = (/&
       0.009523809523809525d0, &
       0.009523809523809525d0, &
       0.009523809523809525d0, &
       0.009523809523809525d0, &
       0.009523809523809525d0, &
       0.009523809523809525d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0, &
       0.028571428571428571d0/)
 
else if(integrationpoints ==  26) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.707106781186547573d0, &
       0.707106781186547573d0, &
      -0.707106781186547573d0, &
      -0.707106781186547573d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0/)
 
weight = (/&
       0.047619047619047623d0, &
       0.047619047619047623d0, &
       0.047619047619047623d0, &
       0.047619047619047623d0, &
       0.047619047619047623d0, &
       0.047619047619047623d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.038095238095238099d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0, &
       0.032142857142857140d0/)
 
else if(integrationpoints ==  14) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
       0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0, &
      -0.577350269189625731d0/)
 
weight = (/&
       0.066666666666666666d0, &
       0.066666666666666666d0, &
       0.066666666666666666d0, &
       0.066666666666666666d0, &
       0.066666666666666666d0, &
       0.066666666666666666d0, &
       0.074999999999999997d0, &
       0.074999999999999997d0, &
       0.074999999999999997d0, &
       0.074999999999999997d0, &
       0.074999999999999997d0, &
       0.074999999999999997d0, &
       0.074999999999999997d0, &
       0.074999999999999997d0/)
 
else if(integrationpoints ==   6) then
r1 = (/&
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0/)
r2 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0/)
 
r3 = (/&
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       0.000000000000000000d0, &
       1.000000000000000000d0, &
      -1.000000000000000000d0/)
 
weight = (/&
       0.166666666666666713d0, &
       0.166666666666666713d0, &
       0.166666666666666713d0, &
       0.166666666666666713d0, &
       0.166666666666666713d0, &
       0.166666666666666713d0/)
 
else
write(*,*) "Error in Lebedev.f90: invalid number of Lebedev integration points."
write(*,*) "Only the following values are allowed: 6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302, 350, 434, 4802" 
end if
 


if (j > integrationpoints) then
   write(*,*) 'IN < LEBEDEV >: j = ',j,' > integrationpoints = ',integrationpoints
   stop
end if

r1_current=r1(j)
r2_current=r2(j)
r3_current=r3(j)
weight_current=weight(j)*4*Pi ! normalise to 4Pi (surface of the unit sphere) instead of 1
end subroutine lebedevpoint

end module Lebedev

